#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot.enums import FairnessMetricsSet


@pytest.fixture
def fairness_insights_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/fairnessInsights/".format(
        project_id, model_id
    )


@pytest.fixture
def data_disparity_insights_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/dataDisparityInsights/".format(
        project_id, model_id
    )


@pytest.fixture
def cross_class_accuracy_scores_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/crossClassAccuracyScores/".format(
        project_id, model_id
    )


@pytest.fixture
def fairness_insights_job_creation_response(fairness_insights_url, job_url):
    responses.add(
        responses.POST,
        fairness_insights_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def data_disparity_insights_job_creation_response(data_disparity_insights_url, job_url):
    responses.add(
        responses.POST,
        data_disparity_insights_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def cross_class_accuracy_scores_job_creation_response(cross_class_accuracy_scores_url, job_url):
    responses.add(
        responses.POST,
        cross_class_accuracy_scores_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


def fairness_insights_server_data(model_id, fairness_metric):
    return {
        "count": 1,
        "next": None,
        "previous": None,
        "data": [
            {
                "modelId": "".format(model_id),
                "fairnessMetric": fairness_metric,
                "fairnessThreshold": 0.8,
                "predictionThreshold": 0.5,
                "protectedFeature": "race",
                "perClassFairness": [
                    {
                        "className": "Amer-Indian-Eskimo",
                        "value": 0.35355038390571986,
                        "absoluteValue": 0.07547169811320754,
                        "entriesCount": 53,
                        "isStatisticallySignificant": False,
                    },
                    {
                        "className": "Asian-Pac-Islander",
                        "value": 1,
                        "absoluteValue": 0.23417721518987342,
                        "entriesCount": 158,
                        "isStatisticallySignificant": False,
                    },
                    {
                        "className": "White",
                        "value": 1.0,
                        "absoluteValue": 0.21346801346801347,
                        "entriesCount": 4455,
                        "isStatisticallySignificant": True,
                    },
                ],
            },
        ],
        "totalCount": 1,
    }


@pytest.fixture
def feature():
    return "race"


@pytest.fixture
def class_name1():
    return "Amer-Indian-Eskimo"


@pytest.fixture
def class_name2():
    return "Black"


@pytest.fixture
def data_disparity_insights_server_data(feature, class_name1, class_name2):
    return {
        "count": 2,
        "next": None,
        "previous": None,
        "totalCount": 2,
        "data": {
            "protectedFeature": feature,
            "metric": "psi",
            "values": [{"label": class_name1, "count": 53}, {"label": class_name2, "count": 511}],
            "features": [
                {
                    "name": "salary",
                    "disparityScore": 0.019521590824952226,
                    "featureImpact": 1,
                    "status": "Healthy",
                    "detailsHistogram": [
                        {
                            "bin": ">50K",
                            "bars": [
                                {"label": class_name1, "value": 0.09433962264150944},
                                {"label": class_name2, "value": 0.13894324853228962},
                            ],
                        },
                    ],
                },
                {
                    "name": "capital_gain",
                    "disparityScore": 0.03430540040461845,
                    "featureImpact": 1.0,
                    "status": "Healthy",
                    "detailsHistogram": [
                        {
                            "bin": "[0.0, 9999.9)",
                            "bars": [
                                {"label": class_name1, "value": 0.9811320754716981},
                                {"label": class_name2, "value": 0.9804305283757339},
                            ],
                        },
                        {
                            "bin": "[9999.9, 19999.8)",
                            "bars": [
                                {"label": class_name1, "value": 0.018867924528301886},
                                {"label": class_name2, "value": 0.009784735812133072},
                            ],
                        },
                    ],
                },
            ],
        },
    }


@pytest.fixture
def cross_class_accuracy_scores_server_data(model_id, feature, class_name1, class_name2):
    return {
        "count": 1,
        "next": None,
        "previous": None,
        "data": [
            {
                "feature": feature,
                "modelId": model_id,
                "predictionThreshold": 0.5,
                "perClassAccuracyScores": [
                    {
                        "className": class_name1,
                        "metrics": [
                            {"metric": "LogLoss", "value": 0.15817195767047793},
                            {"metric": "f1", "value": 0.7175572519083969},
                            {"metric": "AUC", "value": 0.9645006402048656},
                            {"metric": "accuracy", "value": 0.9275929549902152},
                        ],
                    },
                    {
                        "className": class_name2,
                        "metrics": [
                            {"metric": "LogLoss", "value": 0.10583976310412195},
                            {"metric": "f1", "value": 0.8888888888888888},
                            {"metric": "AUC", "value": 1.0},
                            {"metric": "accuracy", "value": 0.9811320754716981},
                        ],
                    },
                ],
            },
        ],
        "totalCount": 1,
    }


@pytest.fixture
def fairness_insights_response_equal_parity(fairness_insights_url, model_id):
    body = json.dumps(fairness_insights_server_data(model_id, FairnessMetricsSet.EQUAL_PARITY))
    responses.add(
        responses.GET,
        "{}?offset=0&limit=100&fairnessMetricsSet={}".format(
            fairness_insights_url, FairnessMetricsSet.EQUAL_PARITY
        ),
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def data_disparity_insights_response(
    data_disparity_insights_url,
    data_disparity_insights_server_data,
    feature,
    class_name1,
    class_name2,
):
    body = json.dumps(data_disparity_insights_server_data)
    responses.add(
        responses.GET,
        "{}?feature={}&className1={}&className2={}".format(
            data_disparity_insights_url, feature, class_name1, class_name2
        ),
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def cross_class_accuracy_scores_response(
    cross_class_accuracy_scores_url, cross_class_accuracy_scores_server_data,
):
    body = json.dumps(cross_class_accuracy_scores_server_data)
    responses.add(
        responses.GET,
        cross_class_accuracy_scores_url,
        status=200,
        content_type="application/json",
        body=body,
    )


@responses.activate
@pytest.mark.usefixtures(
    "fairness_insights_job_creation_response", "fairness_insights_response_equal_parity",
)
def test_request_fairness_insights(one_model, job_id):
    assert one_model.request_fairness_insights() == job_id

    fairness_metrics_set = FairnessMetricsSet.EQUAL_PARITY
    assert one_model.request_fairness_insights(fairness_metrics_set) == job_id

    fairness_insights = one_model.get_fairness_insights(fairness_metrics_set=fairness_metrics_set)
    assert fairness_insights["data"][0]["fairnessMetric"] == fairness_metrics_set


@responses.activate
@pytest.mark.usefixtures(
    "data_disparity_insights_job_creation_response",
    "data_disparity_insights_response",
    "data_disparity_insights_server_data",
)
def test_request_data_disparity_insights(one_model, job_id, feature, class_name1, class_name2):
    assert (
        one_model.request_data_disparity_insights(
            feature=feature, compared_class_names=[class_name1, class_name2]
        )
        == job_id
    )

    data_disparity_insights = one_model.get_data_disparity_insights(
        feature, class_name1, class_name2
    )
    assert data_disparity_insights["data"]["protectedFeature"] == feature


@responses.activate
@pytest.mark.usefixtures(
    "cross_class_accuracy_scores_job_creation_response",
    "cross_class_accuracy_scores_response",
    "cross_class_accuracy_scores_server_data",
)
def test_cross_class_accuracy_scores(one_model, job_id, feature):
    assert one_model.request_cross_class_accuracy_scores() == job_id

    data_disparity_insights = one_model.get_cross_class_accuracy_scores()
    assert data_disparity_insights["data"][0]["feature"] == feature



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from copy import copy, deepcopy
import json
from tempfile import NamedTemporaryFile

import pytest
import responses
import six

from datarobot import CustomInferenceModel, TARGET_TYPE
from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE
from datarobot.utils import underscorize
from tests.model.utils import assert_custom_model_version


@pytest.fixture
def mocked_inference_model(mocked_models):
    return mocked_models["data"][1]


@pytest.fixture
def mocked_inference_model_with_resources(mocked_models_with_resources):
    return deepcopy(mocked_models_with_resources["data"][1])


@pytest.fixture
def mocked_multiclass_inference_model(mocked_multiclass_models):
    return deepcopy(mocked_multiclass_models["data"][1])


@pytest.fixture
def mocked_unstructured_inference_model(mocked_unstructured_models):
    return deepcopy(mocked_unstructured_models["data"][1])


@pytest.fixture
def mocked_anomaly_inference_model(mocked_anomaly_models):
    return deepcopy(mocked_anomaly_models["data"][1])


@pytest.fixture
def mocked_inference_model_no_latest_version(mocked_inference_model):
    model = copy(mocked_inference_model)
    model["latestVersion"] = None
    return model


@pytest.fixture
def mocked_inference_model_with_deprecated_target_support(mocked_inference_model):
    model = deepcopy(mocked_inference_model)
    if model["targetType"] == TARGET_TYPE.BINARY:
        model["supportsBinaryClassification"] = True
    elif model["targetType"] == TARGET_TYPE.REGRESSION:
        model["supportsRegression"] = True
    return model


def assert_custom_model_common(model, model_json):
    assert model.id == model_json["id"]
    assert model.name == model_json["name"]
    assert model.description == model_json["description"]

    for expected_target_type in [
        TARGET_TYPE.BINARY,
        TARGET_TYPE.REGRESSION,
        TARGET_TYPE.MULTICLASS,
        TARGET_TYPE.UNSTRUCTURED,
        TARGET_TYPE.ANOMALY,
    ]:
        assert (model.target_type == expected_target_type) == (
            model_json["targetType"] == expected_target_type
        )

    if model_json["latestVersion"]:
        assert_custom_model_version(model.latest_version, model_json["latestVersion"])
    else:
        assert model.latest_version is None
    assert model.deployments_count == model_json["deploymentsCount"]
    assert model.created_by == model_json["createdBy"]
    assert model.updated_at == model_json["updated"]
    assert model.created_at == model_json["created"]


def assert_inference_model(model, model_json):
    assert_custom_model_common(model, model_json)
    assert model.language == model_json["language"]
    if model.target_type in CUSTOM_MODEL_TARGET_TYPE.REQUIRES_TARGET_NAME:
        assert model.target_name == model_json["targetName"]
    assert model.training_dataset_id == model_json["trainingDatasetId"]
    assert model.training_dataset_version_id == model_json["trainingDatasetVersionId"]
    assert model.training_data_file_name == model_json["trainingDataFileName"]
    assert (
        model.training_data_assignment_in_progress == model_json["trainingDataAssignmentInProgress"]
    )
    assert model.positive_class_label == model_json.get("positiveClassLabel")
    assert model.negative_class_label == model_json.get("negativeClassLabel")
    assert model.prediction_threshold == model_json.get("predictionThreshold")
    assert model.class_labels == model_json.get("classLabels")
    assert model.training_data_partition_column == model_json["trainingDataPartitionColumn"]
    assert model.network_egress_policy == model_json["networkEgressPolicy"]
    assert model.maximum_memory == model_json["maximumMemory"]
    assert model.replicas == model_json["replicas"]


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


class TestCustomInferenceModel(object):
    @pytest.mark.parametrize(
        "model_data", ["mocked_inference_model", "mocked_inference_model_no_latest_version"]
    )
    def test_from_server_data(self, request, model_data):
        model_data = request.getfixturevalue(model_data)
        model = CustomInferenceModel.from_server_data(model_data)
        assert_inference_model(model, model_data)

    @responses.activate
    def test_get_version(self, mocked_inference_model, make_models_url):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_list_versions(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url() + "?customModelType=inference"
        mock_get_response(
            url,
            {
                "count": 1,
                "totalCount": 1,
                "next": None,
                "previous": None,
                "data": [mocked_inference_model],
            },
        )

        # act
        models = CustomInferenceModel.list()

        # assert
        assert len(models) == 1
        assert_inference_model(models[0], mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    @pytest.mark.parametrize("labels_as_files", [True, False])
    def test_update(self, make_models_url, mocked_inference_model, labels_as_files):
        # arrange
        patch = {
            "description": "xx",
            "predictionThreshold": 0.1,
            "classLabels": ["hot dog", "burrito", "hoagie", "reuben"],
        }

        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update(patch)

        responses.add(
            responses.PATCH,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        with NamedTemporaryFile() as f:
            args = deepcopy(patch)
            if labels_as_files:
                f.write("\n".join(args.pop("classLabels")).encode("utf-8"))
                f.flush()
                args["classLabelsFile"] = f.name

            model.update(**{underscorize(k): v for k, v in args.items()})

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == url
        assert responses.calls[1].request.body == json.dumps(patch).encode()

    @responses.activate
    def test_refresh(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update({"description": "xx", "predictionThreshold": 0.1})

        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.refresh()

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_delete(self, make_models_url, mocked_inference_model):
        url = make_models_url(mocked_inference_model["id"])

        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.DELETE, url, status=204, content_type="application/json",
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.delete()

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "DELETE"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_download(self, make_models_url, mocked_inference_model, tmpdir):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.GET,
            url + "download/",
            status=200,
            content_type="application/json",
            body=b"content",
        )

        downloaded_file = tmpdir.mkdir("sub").join("download")
        downloaded_file_path = str(downloaded_file)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.download_latest_version(downloaded_file_path)

        # assert
        assert six.ensure_binary(downloaded_file.read()) == b"content"

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url + "download/"

    @responses.activate
    def test_assign_training_data_non_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        mocked_inference_model["trainingDataAssignmentInProgress"] = True
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae", max_wait=None)

        assert model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(mocked_inference_model["id"])
        assert responses.calls[1].request.method == "PATCH"
        assert (
            responses.calls[1].request.url
            == make_models_url(mocked_inference_model["id"]) + "trainingData/"
        )
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == make_models_url(mocked_inference_model["id"])

    @responses.activate
    def test_assign_training_data_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        responses.add(
            responses.GET,
            status_url,
            status=303,
            content_type="application/json",
            headers={"Location": status_url},
        )
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae")

        assert not model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(mocked_inference_model["id"])
        assert responses.calls[1].request.method == "PATCH"
        assert (
            responses.calls[1].request.url
            == make_models_url(mocked_inference_model["id"]) + "trainingData/"
        )
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == status_url
        assert responses.calls[3].request.method == "GET"
        assert responses.calls[3].request.url == make_models_url(mocked_inference_model["id"])

    @responses.activate
    @pytest.mark.parametrize(
        "model_data",
        [
            "mocked_inference_model",
            "mocked_inference_model_no_latest_version",
            "mocked_inference_model_with_resources",
            "mocked_inference_model_with_deprecated_target_support",
        ],
    )
    def test_create(self, request, make_models_url, model_data):
        model_data = request.getfixturevalue(model_data)
        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        model = CustomInferenceModel.create(
            model_data["name"],
            TARGET_TYPE.REGRESSION,
            model_data["targetName"],
            network_egress_policy=model_data["networkEgressPolicy"],
            maximum_memory=model_data["maximumMemory"],
            replicas=model_data["replicas"],
        )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()

    def test_create_model_without_target_name(self, mocked_inference_model):
        model_data = mocked_inference_model
        with pytest.raises(
            ValueError,
            match="target_name is required for custom models with target type Regression",
        ):
            CustomInferenceModel.create(
                model_data["name"],
                TARGET_TYPE.REGRESSION,
                network_egress_policy=model_data["networkEgressPolicy"],
                maximum_memory=model_data["maximumMemory"],
                replicas=model_data["replicas"],
            )

    @responses.activate
    @pytest.mark.parametrize("labels_as_files", [True, False])
    def test_create_with_multiclass_labels(
        self, mocked_multiclass_inference_model, make_models_url, labels_as_files
    ):
        model_data = mocked_multiclass_inference_model
        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        if labels_as_files:
            with NamedTemporaryFile() as f:
                f.write("\n".join(model_data["classLabels"]).encode("utf-8"))
                f.flush()
                model = CustomInferenceModel.create(
                    model_data["name"],
                    TARGET_TYPE.MULTICLASS,
                    model_data["targetName"],
                    class_labels_file=f.name,
                    network_egress_policy=model_data["networkEgressPolicy"],
                    maximum_memory=model_data["maximumMemory"],
                    replicas=model_data["replicas"],
                )
        else:
            model = CustomInferenceModel.create(
                model_data["name"],
                TARGET_TYPE.MULTICLASS,
                model_data["targetName"],
                class_labels=model_data["classLabels"],
                network_egress_policy=model_data["networkEgressPolicy"],
                maximum_memory=model_data["maximumMemory"],
                replicas=model_data["replicas"],
            )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()
        assert (
            json.loads(responses.calls[0].request.body)["classLabels"] == model_data["classLabels"]
        )

    @responses.activate
    @pytest.mark.parametrize(
        "inference_model, target_type",
        [
            ["mocked_unstructured_inference_model", TARGET_TYPE.UNSTRUCTURED],
            ["mocked_anomaly_inference_model", TARGET_TYPE.ANOMALY],
        ],
    )
    def test_create_no_target_model(self, request, inference_model, target_type, make_models_url):
        model_data = request.getfixturevalue(inference_model)

        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        model = CustomInferenceModel.create(
            model_data["name"],
            target_type,
            network_egress_policy=model_data["networkEgressPolicy"],
            maximum_memory=model_data["maximumMemory"],
            replicas=model_data["replicas"],
        )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()

    @responses.activate
    def test_create_copy(self, make_models_url, mocked_inference_model):
        custom_model_to_copy = "5ea6bbc7402403321d4e1fad"

        url = make_models_url() + "fromCustomModel/"
        responses.add(
            responses.POST,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.copy_custom_model(custom_model_to_copy)
        assert_inference_model(model, mocked_inference_model)

        req = responses.calls[0].request
        assert req.method == "POST"
        assert req.url == url
        assert json.loads(req.body) == {"customModelId": custom_model_to_copy}



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
import responses

from datarobot import Client
from datarobot.client import get_client


def test_missing_token_in_env_fails(clean_env):
    clean_env.setenv("DATAROBOT_ENDPOINT", "https://home.com")
    with pytest.raises(ValueError):
        get_client()


def test_missing_endpoint_in_config_fails():
    with pytest.raises(ValueError):
        Client(token="a-token")


def test_missing_token_in_config_fails():
    with pytest.raises(ValueError):
        Client(endpoint="https://host_name.com")


@pytest.mark.usefixtures("no_client_version_check")
def test_file_config_with_connect_timeout(
    clean_user_config, default_config_token, default_config_endpoint
):
    fcontent = "endpoint: {}\ntoken: {}\nconnect_timeout: 23".format(
        default_config_endpoint, default_config_token
    )
    clean_user_config.write(fcontent)
    client = get_client()
    assert client.connect_timeout == 23


@pytest.mark.usefixtures("no_client_version_check")
def test_code_config_with_connect_timeout(code_config_token, code_config_endpoint):
    Client(token=code_config_token, endpoint=code_config_endpoint, connect_timeout=24)
    c = get_client()
    assert c.connect_timeout == 24


@responses.activate
def test_bad_token_message():
    endpoint = "https://host_name.com/api/v2"
    responses.add(
        responses.GET, endpoint + "/version/", status=401,
    )
    with pytest.raises(UserWarning) as userwarning:
        Client(token="not correct", endpoint=endpoint)
    assert str(userwarning.value) == (
        "Unable to authenticate to the server - are you sure the "
        'provided token of "not correct" and endpoint of "https://'
        'host_name.com/api/v2" are correct? Note that if you access '
        "the DataRobot webapp at `https://app.datarobot.com`, then "
        "the correct endpoint to specify would be "
        "`https://app.datarobot.com/api/v2`."
    )


@responses.activate
def test_bad_endpoint_message():
    endpoint = "https://host_name.com/api/v16"
    responses.add(
        responses.GET, endpoint + "/version/", status=404,
    )
    with pytest.raises(UserWarning) as userwarning:
        Client(token="not correct", endpoint=endpoint)
    assert str(userwarning.value) == (
        "Error retrieving a version string from the server. Server "
        "did not reply with an API version. This may indicate the "
        "endpoint parameter `https://host_name.com/api/v16` is "
        "incorrect, or that the server there is incompatible with "
        "this version of the DataRobot client package. Note that if "
        "you access the DataRobot webapp at `https://app.datarobot"
        ".com`, then the correct endpoint to specify would be "
        "`https://app.datarobot.com/api/v2`."
    )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import OrderedDict

import mock
import pytest
import responses

from datarobot.enums import VARIABLE_TYPE_TRANSFORM
from tests.utils import request_body_to_json


@pytest.fixture
def batch_type_transform_create_url(unittest_endpoint, project_id):
    return "{}/projects/{}/batchTypeTransformFeatures/".format(unittest_endpoint, project_id)


@pytest.fixture
def batch_type_transform_create_response(batch_type_transform_create_url, async_url):
    responses.add(
        responses.POST,
        batch_type_transform_create_url,
        status=202,
        body="",
        adding_headers={"Location": async_url},
    )
    responses.add(
        responses.GET, async_url, status=303, body="", adding_headers={"Location": "/features/"}
    )


@pytest.mark.parametrize(
    "exception, patch",
    [(TypeError, {"parent_names": []}), (ValueError, {"variable_type": "wrong"})],
)
def test_batch_var_type_transform_arguments_validation(project, exception, patch):
    args = OrderedDict(
        {
            "parent_names": ["f1"],
            "variable_type": VARIABLE_TYPE_TRANSFORM.TEXT,
            "prefix": "prefix",
            "suffix": "suffix",
        }
    )

    args.update(patch)

    with pytest.raises(exception):
        project.batch_features_type_transform(
            parent_names=args["parent_names"],
            variable_type=args["variable_type"],
            prefix=args["prefix"],
            suffix=args["suffix"],
        )


@responses.activate
@pytest.mark.usefixtures("batch_type_transform_create_response")
def test_batch_var_type_transform(project):
    args = [
        ["f1", "f2"],
        VARIABLE_TYPE_TRANSFORM.NUMERIC,
        "prefix",
    ]

    with mock.patch.object(project, "get_features") as get_features_mock:
        features = project.batch_features_type_transform(*args)
        assert get_features_mock.called_once
        assert get_features_mock.return_value is features

    payload = request_body_to_json(responses.calls[0].request)

    assert payload["parentNames"] == ["f1", "f2"]
    assert payload["variableType"] == VARIABLE_TYPE_TRANSFORM.NUMERIC
    assert payload["prefix"] == "prefix"

    assert "suffix" not in payload


@responses.activate
@pytest.mark.usefixtures("batch_type_transform_create_response")
@pytest.mark.usefixtures("known_warning")
def test_batch_var_type_transform_categorical(project):
    args = [
        ["f1", "f2"],
        VARIABLE_TYPE_TRANSFORM.CATEGORICAL,
        "prefix",
    ]

    with mock.patch.object(project, "get_features") as get_features_mock:
        features = project.batch_features_type_transform(*args)
        assert get_features_mock.called_once
        assert get_features_mock.return_value is features

    payload = request_body_to_json(responses.calls[0].request)

    assert payload["parentNames"] == ["f1", "f2"]
    assert payload["variableType"] == VARIABLE_TYPE_TRANSFORM.CATEGORICAL
    assert payload["prefix"] == "prefix"

    assert "suffix" not in payload



# coding: utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime
import json

import dateutil
import pytest
import responses

from datarobot.enums import JOB_TYPE
from datarobot.models.model import PrimeModel
from datarobot.models.prime_file import PrimeFile
from datarobot.models.ruleset import Ruleset
from tests.utils import assert_equal_py2, assert_equal_py3


@responses.activate
def test_get_prime_model(
    project_url, project_id, prime_model_id, prime_model_server_data, ruleset_with_model_server_data
):
    responses.add(
        responses.GET,
        "{}primeModels/{}/".format(project_url, prime_model_id),
        body=json.dumps(prime_model_server_data),
    )
    prime_model = PrimeModel.get(project_id, prime_model_id)

    assert prime_model.id == prime_model_server_data["id"]
    assert prime_model.model_type == prime_model_server_data["modelType"]
    assert prime_model.parent_model_id == ruleset_with_model_server_data["parentModelId"]
    assert prime_model.ruleset.score == ruleset_with_model_server_data["score"]


@responses.activate
@pytest.mark.usefixtures("client", "approximate_job_creation_response")
def test_approximate(one_model, job_url, approximate_job_running_server_data):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(approximate_job_running_server_data),
        content_type="application/json",
    )
    approx_job = one_model.request_approximation()
    assert approx_job.job_type == JOB_TYPE.PRIME_RULESETS


@pytest.fixture
def datetime_prime_data(datetime_model_data, prime_data):
    modified_data = dict(prime_data)
    modified_data.update(datetime_model_data)
    modified_data.pop("sample_pct")
    return modified_data


def test_datetime_prime_model(datetime_prime_data):
    mod = PrimeModel.from_data(datetime_prime_data)
    dt = mod.training_start_date
    assert dt.tzinfo == dateutil.tz.tzutc()
    assert isinstance(mod.training_end_date, datetime)
    assert mod.training_row_count is None
    assert mod.training_duration is None


@responses.activate
@pytest.mark.usefixtures(
    "client", "approximate_job_creation_response", "approximate_completed_response"
)
def test_retrieve_approximate_result(one_model, prime_model_id):
    approx_job = one_model.request_approximation()
    rulesets = approx_job.get_result_when_complete()
    assert rulesets
    assert {prime_model_id, None} == {ruleset.model_id for ruleset in rulesets}
    for ruleset in rulesets:
        assert isinstance(ruleset, Ruleset)
        assert ruleset.parent_model_id == one_model.id


@responses.activate
def test_check_primeable(project_url, one_model):
    response_data = {"canMakePrime": True, "message": "OK", "messageId": 0}
    url = "{}models/{}/primeInfo/".format(project_url, one_model.id)
    responses.add(
        responses.GET, url, body=json.dumps(response_data), content_type="application/json"
    )
    eligibility = one_model.get_prime_eligibility()
    assert eligibility == {"can_make_prime": True, "message": "OK"}


@responses.activate
def test_check_primeable_future_proof(project_url, one_model):
    response_data = {"canMakePrime": True, "message": "OK", "messageId": 0, "future": "new"}
    url = "{}models/{}/primeInfo/".format(project_url, one_model.id)
    responses.add(
        responses.GET, url, body=json.dumps(response_data), content_type="application/json"
    )
    eligibility = one_model.get_prime_eligibility()
    assert eligibility == {"can_make_prime": True, "message": "OK"}


@responses.activate
def test_list_rulesets(
    project_url,
    model_id,
    prime_model_id,
    one_model,
    ruleset_with_model_server_data,
    ruleset_without_model_server_data,
):
    response_data = [ruleset_with_model_server_data, ruleset_without_model_server_data]
    responses.add(
        responses.GET,
        "{}models/{}/primeRulesets/".format(project_url, model_id),
        body=json.dumps(response_data),
    )
    rulesets = one_model.get_rulesets()

    assert len(rulesets) == 2
    ruleset_models = {ruleset.model_id for ruleset in rulesets}
    assert {prime_model_id, None} == ruleset_models


def test_future_proof(prime_model_server_data):
    future_data = dict(prime_model_server_data, future="new")
    PrimeModel.from_server_data(future_data)


@responses.activate
@pytest.mark.usefixtures("client", "prime_validation_job_creation_response")
def test_validate(prime_model, job_url, prime_validation_job_running_server_data):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(prime_validation_job_running_server_data),
        content_type="application/json",
    )
    validate_job = prime_model.request_download_validation("Python")
    assert validate_job.job_type == JOB_TYPE.PRIME_VALIDATION


@responses.activate
@pytest.mark.usefixtures(
    "client", "prime_validation_job_creation_response", "prime_validation_job_completed_response"
)
def test_retrieve_validate_result(prime_model):
    validate_job = prime_model.request_download_validation("Python")
    prime_file = validate_job.get_result_when_complete()
    assert isinstance(prime_file, PrimeFile)
    assert prime_file.parent_model_id == prime_model.parent_model_id
    assert prime_file.language == "Python"


class TestRepr(object):
    """Tests of method PrimeModel.__repr__"""

    @pytest.fixture
    def kwargs(self):
        """Required parameters of PrimeModel that don't affect the tests"""
        return {
            "ruleset_id": 1,
            "rule_count": 1,
            "score": 1.0,
            "parent_model_id": "58668f2fbf36cd0183a0617a",
            "project_id": "58668f44bf36cd0183a0617b",
        }

    def test_model_type_str(self, kwargs):
        # when both model_type and id are set
        frozen_model = PrimeModel(id="58668cb5bf36cd77a5d8056c", model_type="model_type", **kwargs)
        # then repr contains only model_type
        assert repr(frozen_model) == "PrimeModel('model_type')"

    def test_model_type_unicode(self, kwargs):
        # when both model_type and id are set
        # and model_type contains non-ascii characters
        frozen_model = PrimeModel(id="58668cb5bf36cd77a5d8056c", model_type=u"тайп", **kwargs)
        # then repr is an ascii string containing only model_type
        assert_equal_py2(repr(frozen_model), "PrimeModel(u'\\u0442\\u0430\\u0439\\u043f')")
        assert_equal_py3(repr(frozen_model), "PrimeModel('тайп')")

    def test_id(self, kwargs):
        # when id is set and model_type is missing
        frozen_model = PrimeModel(id="58668cb5bf36cd77a5d8056c", **kwargs)
        # then repr contains id
        assert repr(frozen_model) == "PrimeModel('58668cb5bf36cd77a5d8056c')"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import unittest

from datarobot.helpers.feature_discovery import DatasetDefinition, Relationship, SecondaryDataset


class TestFeatureDiscoveryHelper(unittest.TestCase):
    def test_dataset_definition_to_payload(self):
        dataset_definition = DatasetDefinition(
            identifier="profile",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
        )

        dd_payload = dataset_definition.to_payload()
        assert dd_payload["identifier"] == "profile"
        assert dd_payload["catalogId"] == "5ec4aec1f072bc028e3471ae"
        assert dd_payload["catalogVersionId"] == "5ec4aec2f072bc028e3471b1"

        dataset_definition = DatasetDefinition(
            identifier="transaction",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
            feature_list_id="5ec4aec2f072bc028e3471b5",
            primary_temporal_key="Date",
        )
        dd_payload = dataset_definition.to_payload()
        assert dd_payload["identifier"] == "transaction"
        assert dd_payload["catalogId"] == "5ec4aec1f072bc028e3471ae"
        assert dd_payload["catalogVersionId"] == "5ec4aec2f072bc028e3471b1"
        assert dd_payload["featureListId"] == "5ec4aec2f072bc028e3471b5"
        assert dd_payload["primaryTemporalKey"] == "Date"

    def test_dataset_definition_to_dict(self):
        dataset_definition = DatasetDefinition(
            identifier="profile",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
        )

        dd_dict = dataset_definition.to_dict()
        assert dd_dict["identifier"] == "profile"
        assert dd_dict["catalog_id"] == "5ec4aec1f072bc028e3471ae"
        assert dd_dict["catalog_version_id"] == "5ec4aec2f072bc028e3471b1"
        assert dd_dict["snapshot_policy"] == "latest"
        assert dd_dict["feature_list_id"] is None
        assert dd_dict["primary_temporal_key"] is None

        dataset_definition = DatasetDefinition(
            identifier="transaction",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
            feature_list_id="5ec4aec2f072bc028e3471b5",
            primary_temporal_key="Date",
        )
        dd_dict = dataset_definition.to_dict()
        assert dd_dict["identifier"] == "transaction"
        assert dd_dict["catalog_id"] == "5ec4aec1f072bc028e3471ae"
        assert dd_dict["catalog_version_id"] == "5ec4aec2f072bc028e3471b1"
        assert dd_dict["snapshot_policy"] == "latest"
        assert dd_dict["feature_list_id"] == "5ec4aec2f072bc028e3471b5"
        assert dd_dict["primary_temporal_key"] == "Date"

    def test_relationship_to_payload(self):
        relationship = Relationship(
            dataset1_identifier="profile",
            dataset2_identifier="transaction",
            dataset1_keys=["CustomerID"],
            dataset2_keys=["CustomerID"],
        )
        rel_payload = relationship.to_payload()
        assert rel_payload["dataset1Identifier"] == "profile"
        assert rel_payload["dataset2Identifier"] == "transaction"
        assert rel_payload["dataset1Keys"] == ["CustomerID"]
        assert rel_payload["dataset2Keys"] == ["CustomerID"]
        assert "featureDerivationWindowStart" not in rel_payload
        assert "featureDerivationWindowEnd" not in rel_payload
        assert "featureDerivationWindowTimeUnit" not in rel_payload
        assert "predictionPointRounding" not in rel_payload
        assert "predictionPointRoundingTimeUnit" not in rel_payload

        relationship = Relationship(
            dataset2_identifier="profile",
            dataset1_keys=["CustomerID"],
            dataset2_keys=["CustomerID"],
            feature_derivation_window_start=-14,
            feature_derivation_window_end=0,
            feature_derivation_window_time_unit="DAY",
            prediction_point_rounding=0,
            prediction_point_rounding_time_unit="DAY",
        )
        rel_payload = relationship.to_payload()
        assert rel_payload["dataset2Identifier"] == "profile"
        assert rel_payload["dataset1Keys"] == ["CustomerID"]
        assert rel_payload["dataset2Keys"] == ["CustomerID"]
        assert "dataset1Identifier" not in rel_payload
        assert rel_payload["featureDerivationWindowStart"] == -14
        assert rel_payload["featureDerivationWindowEnd"] == 0
        assert rel_payload["featureDerivationWindowTimeUnit"] == "DAY"
        assert rel_payload["predictionPointRounding"] == 0
        assert rel_payload["predictionPointRoundingTimeUnit"] == "DAY"

    def test_relationship_to_dict(self):
        relationship = Relationship(
            dataset1_identifier="profile",
            dataset2_identifier="transaction",
            dataset1_keys=["CustomerID"],
            dataset2_keys=["CustomerID"],
        )
        rel_dict = relationship.to_dict()
        assert rel_dict["dataset1_identifier"] == "profile"
        assert rel_dict["dataset2_identifier"] == "transaction"
        assert rel_dict["dataset1_keys"] == ["CustomerID"]
        assert rel_dict["dataset2_keys"] == ["CustomerID"]
        assert rel_dict["feature_derivation_window_start"] is None
        assert rel_dict["feature_derivation_window_end"] is None
        assert rel_dict["feature_derivation_window_time_unit"] is None
        assert rel_dict["prediction_point_rounding"] is None
        assert rel_dict["prediction_point_rounding_time_unit"] is None

        relationship = Relationship(
            dataset2_identifier="profile",
            dataset1_keys=["CustomerID"],
            dataset2_keys=["CustomerID"],
            feature_derivation_window_start=-14,
            feature_derivation_window_end=0,
            feature_derivation_window_time_unit="DAY",
            prediction_point_rounding=0,
            prediction_point_rounding_time_unit="DAY",
        )
        rel_dict = relationship.to_dict()
        assert rel_dict["dataset2_identifier"] == "profile"
        assert rel_dict["dataset1_keys"] == ["CustomerID"]
        assert rel_dict["dataset2_keys"] == ["CustomerID"]
        assert rel_dict["dataset1_identifier"] is None
        assert rel_dict["feature_derivation_window_start"] == -14
        assert rel_dict["feature_derivation_window_end"] == 0
        assert rel_dict["feature_derivation_window_time_unit"] == "DAY"
        assert rel_dict["prediction_point_rounding"] == 0
        assert rel_dict["prediction_point_rounding_time_unit"] == "DAY"

    def test_secondary_dataset_to_payload(self):
        secondary_dataset = SecondaryDataset(
            identifier="profile",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
        )

        sec_dataset_payload = secondary_dataset.to_payload()
        assert sec_dataset_payload["identifier"] == "profile"
        assert sec_dataset_payload["catalogId"] == "5ec4aec1f072bc028e3471ae"
        assert sec_dataset_payload["catalogVersionId"] == "5ec4aec2f072bc028e3471b1"
        assert sec_dataset_payload["snapshotPolicy"] == "latest"

    def test_secondary_dataset_to_dict(self):
        secondary_dataset = SecondaryDataset(
            identifier="profile",
            catalog_id="5ec4aec1f072bc028e3471ae",
            catalog_version_id="5ec4aec2f072bc028e3471b1",
        )

        sec_dataset_dict = secondary_dataset.to_dict()
        assert sec_dataset_dict["identifier"] == "profile"
        assert sec_dataset_dict["catalog_id"] == "5ec4aec1f072bc028e3471ae"
        assert sec_dataset_dict["catalog_version_id"] == "5ec4aec2f072bc028e3471b1"
        assert sec_dataset_dict["snapshot_policy"] == "latest"



#! /usr/bin/env python3
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""Show image embedding vectors.

The following will open a project, get the first model id where the feature
name matches, and then print out the image id and the embedding vector.
"""
from datarobot.models import Project
from datarobot.models.visualai import ImageEmbedding


def print_vectors(project_name, feature_name):
    project = Project.list(search_params={"project_name": project_name})[0]
    model_id = next(mid for mid, name in ImageEmbedding.models(project.id) if name == feature_name)
    for embed in ImageEmbedding.list(project.id, model_id, feature_name):
        print("{0} [{1:1.6f}, {2:1.6f}]".format(embed.image.id, embed.position_x, embed.position_y))


if __name__ == "__main__":
    project_name = "dataset_2k.zip"
    feature_name = "image"
    print_vectors(project_name, feature_name)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from mock import patch
import pytest
import responses

from datarobot import ExternalMulticlassLiftChart
from datarobot.errors import ClientError
from datarobot.utils import from_api


class TestExternalMulticlassLiftChart(object):
    @pytest.fixture()
    def multiclass_liftchart_classbins_data(self, lift_chart_bins_data):
        return [
            {"targetClass": "classA", "bins": lift_chart_bins_data},
            {"targetClass": "classB", "bins": lift_chart_bins_data},
            {"targetClass": "classC", "bins": lift_chart_bins_data},
        ]

    @pytest.fixture()
    def one_class_liftchart(self, lift_chart_bins_data):
        return {
            "targetClass": "classA",
            "bins": lift_chart_bins_data,
            "datasetId": "dataset_id_1",
        }

    @staticmethod
    def create_response_data(multiclass_liftchart_classbins_data, count):
        return [
            {
                "classBins": multiclass_liftchart_classbins_data,
                "datasetId": "dataset_id_{}".format(i),
            }
            for i in range(count)
        ]

    @pytest.fixture
    def url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/datasetMulticlassLiftCharts/".format(
            project_id, model_id
        )

    def test_instantiation(self, one_class_liftchart):
        lc = ExternalMulticlassLiftChart.from_server_data(one_class_liftchart)
        assert lc.dataset_id == one_class_liftchart["datasetId"]
        assert lc.bins == from_api(one_class_liftchart["bins"])
        assert lc.target_class == from_api(one_class_liftchart["targetClass"])

    def test_future_proof(self, one_class_liftchart):
        data_with_future_keys = dict(one_class_liftchart, new_key="some future lift data")
        data_with_future_keys["bins"][0]["new_key"] = "some future bin data"
        ExternalMulticlassLiftChart.from_server_data(data_with_future_keys)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_multiclass_lift_chart."
        "DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_lift_chart_all_results(
        self, multiclass_liftchart_classbins_data, url, project_id, model_id
    ):
        url_page1 = url + "?limit=10&offset=0"
        url_page2 = url + "?limit=10&offset=10"
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True,
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True,
        )

        lift_charts = ExternalMulticlassLiftChart.list(project_id, model_id, limit=0)
        n_classes = len(multiclass_liftchart_classbins_data)
        assert len(lift_charts) == 11 * n_classes
        for j in range(11):
            for k, class_bin in enumerate(multiclass_liftchart_classbins_data):
                i = j * n_classes + k
                assert isinstance(lift_charts[i], ExternalMulticlassLiftChart)
                assert lift_charts[i].dataset_id == "dataset_id_{}".format(j)
                assert lift_charts[i].bins == from_api(class_bin["bins"])
                assert lift_charts[i].target_class == class_bin["targetClass"]

    @responses.activate
    def test_list_lift_chart_limited(
        self, multiclass_liftchart_classbins_data, lift_chart_bins_data, url, project_id, model_id,
    ):
        url_page = url + "?limit=3&offset=2"
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        lift_charts = ExternalMulticlassLiftChart.list(project_id, model_id, limit=3, offset=2)
        n_classes = len(multiclass_liftchart_classbins_data)
        assert len(lift_charts) == 3 * n_classes
        for j in range(3):
            for k, class_bin in enumerate(multiclass_liftchart_classbins_data):
                i = j * n_classes + k
                assert isinstance(lift_charts[i], ExternalMulticlassLiftChart)
                assert lift_charts[i].dataset_id == "dataset_id_{}".format(j + 2)
                assert lift_charts[i].bins == from_api(class_bin["bins"])
                assert lift_charts[i].target_class == class_bin["targetClass"]

    @responses.activate
    def test_get_lift_chart_by_dataset(
        self, multiclass_liftchart_classbins_data, lift_chart_bins_data, url, project_id, model_id,
    ):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        # chart for each class
        lift_charts = ExternalMulticlassLiftChart.list(project_id, model_id, dataset_id)
        assert len(lift_charts) == len(multiclass_liftchart_classbins_data)
        for chart in lift_charts:
            assert isinstance(chart, ExternalMulticlassLiftChart)
            assert chart.dataset_id == dataset_id
            assert chart.bins == from_api(lift_chart_bins_data)

    @responses.activate
    def test_get_lift_chart_by_target_class(
        self, multiclass_liftchart_classbins_data, lift_chart_bins_data, url, project_id, model_id,
    ):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)

        lift_chart = ExternalMulticlassLiftChart.get(
            project_id, model_id, dataset_id, target_class="classB"
        )
        assert isinstance(lift_chart, ExternalMulticlassLiftChart)
        assert lift_chart.dataset_id == dataset_id
        assert lift_chart.bins == from_api(lift_chart_bins_data)
        assert lift_chart.target_class == "classB"

    @responses.activate
    def test_get_lift_chart_by_target_class_failure(
        self, multiclass_liftchart_classbins_data, lift_chart_bins_data, url, project_id, model_id,
    ):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        all_data = self.create_response_data(multiclass_liftchart_classbins_data, 1)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        with pytest.raises(ClientError, match="Requested multiclass lift chart does not exist."):
            ExternalMulticlassLiftChart.get(
                project_id, model_id, dataset_id, target_class="wrong_class"
            )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import contextlib
import json
import re
import sys
import unittest
import warnings

import responses
from six.moves.urllib.parse import parse_qsl, unquote_plus, urlparse

from datarobot.client import set_client
from datarobot.rest import RESTClientObject


class SDKTestcase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        set_client(None)
        c = RESTClientObject(auth="t-token", endpoint="https://host_name.com")
        set_client(c)

    @classmethod
    def tearDownClass(cls):
        set_client(None)


@contextlib.contextmanager
def warns(*categories):
    """Assert warnings are raised in a block, analogous to pytest.raises."""
    with warnings.catch_warnings(record=True) as messages:
        yield messages
    assert tuple(message.category for message in messages) == categories


def assert_raised_regex(exc_info, regexp):
    """
    Until pytest 2.10 when this gets implemented as a method of exc_info,
    we will provide this assertion.

    This is the exact implementation that has already been merged to pytest

    Parameters
    ----------
    exc_info : pytest.ExceptionInfo
    regexp

    Returns
    -------

    """
    if not re.search(regexp, str(exc_info.value)):
        assert 0, "Pattern '{0!s}' not found in '{1!s}'".format(regexp, exc_info.value)
    return True


def assert_equal_py2(this, that):
    """Assert that arguments are equal on Python 2.

    This function does nothing on Python 3.
    """
    if sys.version_info < (3, 0):
        assert this == that


def assert_equal_py3(this, that, message=""):
    """Assert that arguments are equal on Python 3.

    This function does nothing on Python 2.
    """
    if sys.version_info >= (3, 0):
        assert this == that


def assert_urls_equal(url1, url2):
    """Assert that two URLs are equal taking into account quoting and order or parameters"""

    def get_parts(url):
        parts = urlparse(url)
        _query = frozenset(parse_qsl(parts.query))
        _path = unquote_plus(parts.path)
        parts = parts._replace(query=_query, path=_path)
        return parts

    assert get_parts(url1) == get_parts(url2), "{!r} != {!r}".format(url1, url2)


def request_body_to_json(request):
    """Extract json from body of prepared request from requests library.
    Request body is str in Python 2 and bytes in Python 3"""
    body = request.body
    if isinstance(body, bytes):
        return json.loads(body.decode())
    else:
        return json.loads(body)


def add_response(url, body, method=responses.GET, status=200, content_type="application/json"):
    if content_type == "application/json":
        body = json.dumps(body)
    responses.add(method, url, status=status, content_type=content_type, body=body)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import mock
import pytest
import responses

import datarobot.client
from datarobot.client import get_client


@pytest.yield_fixture
def client_version_mock():
    with mock.patch("datarobot.client._get_client_version") as mv:
        yield mv


@pytest.fixture
def version_endpoint(default_config_endpoint):
    return default_config_endpoint + "/version/"


class TestVersionCheck(object):
    @pytest.fixture
    def server_version_string(self):
        yield "2.3.0"

    @pytest.fixture(autouse=True)
    def server_version_response(self, version_endpoint, server_version_string):
        responses.add(
            responses.GET,
            version_endpoint,
            status=200,
            body=json.dumps({"major": "2", "minor": "3", "versionString": server_version_string}),
        )

    @pytest.yield_fixture(autouse=True)
    def no_global_client(self, default_config):
        with mock.patch("datarobot.client._global_client", None) as gc:
            yield gc

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_major_incompat(self, client_version_mock):
        client_version_mock.return_value = "3.0.0"
        client = get_client()
        assert client is None
        assert datarobot.client._global_client is None

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_minor_incompat(self, client_version_mock):
        client_version_mock.return_value = "2.4.0"
        client = get_client()
        assert client is not None
        assert datarobot.client._global_client is client

    @responses.activate
    @pytest.mark.parametrize("server_version_string", ["2.4.0"])
    def test_server_can_be_ahead_of_client(self, client_version_mock):
        client_version_mock.return_value = "2.3.0"
        client = get_client()
        assert client is not None
        assert datarobot.client._global_client is client

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_client_version_import_is_present(self):
        try:
            client = get_client()
            assert client is not None
            assert datarobot.client._global_client is client
        except NameError:
            pytest.fail("Version import statement is not present in client.py")


class TestServerVersion(object):
    @responses.activate
    @pytest.mark.usefixtures("default_config", "known_warning")
    def test_version_does_not_exist(self, version_endpoint):
        """This situation happens if a user is using a client that is new
        enough to check for version on a server that doesn't have an endpoint
        to supply that version
        """
        responses.add(responses.GET, version_endpoint, status=404, body="")

        get_client()

    @responses.activate
    @pytest.mark.usefixtures("default_config")
    def test_version_does_exist(self, version_endpoint, client_version_mock):
        client_version_mock.return_value = "2.4.0"

        responses.add(
            responses.GET,
            version_endpoint,
            status=200,
            body=json.dumps({"major": "2", "minor": "4", "versionString": "2.4.0"}),
        )

        get_client()


class TestClientServerEndpointProtocol(object):
    @pytest.fixture(autouse=True)
    def version_patch(self):
        with mock.patch("datarobot.client._get_client_version") as gv:
            gv.return_value = "2.4.0"
            yield gv

    @responses.activate
    @pytest.mark.parametrize(
        "default_config_endpoint, expected_endpoint",
        [
            pytest.param("https://user_config.com", "https://user_config.com", id="https"),
            pytest.param("http://user_config.com", "http://user_config.com", id="http"),
        ],
    )
    @pytest.mark.usefixtures("default_config")
    def test_correct_configuration_no_changes(self, version_endpoint, expected_endpoint):
        responses.add(
            responses.GET,
            version_endpoint,
            status=200,
            body=json.dumps({"major": "2", "minor": "4", "versionString": "2.4.0"}),
        )

        client = get_client()
        assert client.endpoint == expected_endpoint

    @responses.activate
    @pytest.mark.usefixtures("default_config")
    def test_http_https_protocol_missmatch(self, version_endpoint):
        """If a user specifies http but the server tells them to upgrade to https,
        we want the user to be made aware of it so they can update their config
        """
        https_version_endpoint = version_endpoint.replace("http://", "https://")
        responses.add(
            responses.GET,
            version_endpoint,
            status=301,
            headers={"Location": https_version_endpoint},
        )
        responses.add(
            responses.GET,
            https_version_endpoint,
            status=200,
            body=json.dumps({"major": "2", "minor": "4", "versionString": "2.4.0"}),
        )
        with pytest.warns(UserWarning):  # Expected: warns user about the http/https mismatch
            client = get_client()
        assert client.endpoint.startswith("https://")

    @responses.activate
    @pytest.mark.usefixtures("default_config")
    def test_protocol_check_no_version_returned(self, version_endpoint):
        """There is code that checks for this scenario explicitly. I have no idea why, as it seems like
        `versionString` has always been present on the server response, and has never been empty.
        Also, there is no one else to ask - I am the one who put in the check.
        """
        https_version_endpoint = version_endpoint.replace("http://", "https://")
        responses.add(
            responses.GET,
            version_endpoint,
            status=301,
            headers={"Location": https_version_endpoint},
        )
        responses.add(
            responses.GET,
            https_version_endpoint,
            status=200,
            body=json.dumps({"major": "2", "minor": "4", "versionString": None}),
        )
        with pytest.warns(UserWarning):  # Expected: warns user about the http/https mismatch
            client = get_client()
        assert client.endpoint.startswith("https://")



# -*- encoding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import json
import unittest

import mock
import pandas as pd
import pytest
import responses

from datarobot import errors, Job, PredictJob, Project
from datarobot.models.predict_job import wait_for_async_predictions
from datarobot.utils import raw_prediction_response_to_dataframe
from tests.test_helpers import fixture_file_path
from tests.utils import assert_raised_regex, SDKTestcase

SAMPLE_BINARY_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Binary",
        u"predictions": [
            {
                u"positiveProbability": 0.9,
                u"prediction": 1.0,
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.1,
                u"prediction": 0.0,
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.8,
                u"prediction": 1.0,
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.8, u"label": u"1.0"},
                    {u"value": 0.2, u"label": u"0.0"},
                ],
            },
        ],
    }
)


SAMPLE_BINARY_FLOAT_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Binary",
        u"predictions": [
            {
                u"positiveProbability": 0.9,
                u"prediction": 1.0,
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.9, u"label": 1.0},
                    {u"value": 0.1, u"label": 0.0},
                ],
            },
            {
                u"positiveProbability": 0.1,
                u"prediction": 0.0,
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.9, u"label": 1.0},
                    {u"value": 0.1, u"label": 0.0},
                ],
            },
            {
                u"positiveProbability": 0.8,
                u"prediction": 1.0,
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.8, u"label": 1.0},
                    {u"value": 0.2, u"label": 0.0},
                ],
            },
        ],
    }
)


SAMPLE_MULTICLASS_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Multiclass",
        u"predictions": [
            {
                u"positiveProbability": None,
                u"prediction": u"class_a",
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.7, u"label": u"class_a"},
                    {u"value": 0.2, u"label": u"class_b"},
                    {u"value": 0.1, u"label": u"class_c"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_a",
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.6, u"label": u"class_a"},
                    {u"value": 0.2, u"label": u"class_b"},
                    {u"value": 0.2, u"label": u"class_c"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_b",
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.1, u"label": u"class_a"},
                    {u"value": 0.8, u"label": u"class_b"},
                    {u"value": 0.1, u"label": u"class_c"},
                ],
            },
        ],
    }
)


UNICODE_SAMPLE_MULTICLASS_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Multiclass",
        u"predictions": [
            {
                u"positiveProbability": None,
                u"prediction": u"class_赤木",
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.7, u"label": u"class_赤木"},
                    {u"value": 0.2, u"label": u"class_エピソード"},
                    {u"value": 0.1, u"label": u"class_外"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_赤木",
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.6, u"label": u"class_赤木"},
                    {u"value": 0.2, u"label": u"class_エピソード"},
                    {u"value": 0.2, u"label": u"class_外"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_エピソード",
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.1, u"label": u"class_赤木"},
                    {u"value": 0.8, u"label": u"class_エピソード"},
                    {u"value": 0.1, u"label": u"class_外"},
                ],
            },
        ],
    }
)


SAMPLE_REGRESSION_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": None,
        u"task": u"Regression",
        u"predictions": [
            {u"positiveProbability": None, u"prediction": 32.0, u"rowId": 0},
            {u"positiveProbability": None, u"prediction": 100.0, u"rowId": 1},
            {u"positiveProbability": None, u"prediction": 212.0, u"rowId": 2},
        ],
    }
)


SAMPLE_REGRESSION_PREDICTION_WITH_SHAP_EXPL_RETURN = json.dumps(
    {
        "shapBaseValue": [31108.28125],
        "task": "Regression",
        "positiveClass": None,
        "predictions": [
            {
                "positiveProbability": None,
                "predictionExplanationMetadata": {"shapRemainingTotal": -6969.941378326501},
                "prediction": 27229.166666666668,
                "rowId": 0,
                "predictionExplanations": [
                    {
                        "featureValue": "2004",
                        "strength": 3090.8262156678165,
                        "feature": "YearMade",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "999089",
                        "strength": -2788.4580943655205,
                        "feature": "MachineID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "2006",
                        "strength": -2286.181442643504,
                        "feature": "saledate (Year)",
                        "label": "SalePrice",
                    },
                ],
            },
            {
                "positiveProbability": None,
                "predictionExplanationMetadata": {"shapRemainingTotal": 13439.981135022792},
                "prediction": 60583.333333333336,
                "rowId": 1,
                "predictionExplanations": [
                    {
                        "featureValue": "117657",
                        "strength": 16035.069717943523,
                        "feature": "MachineID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "77",
                        "strength": 10105.427194858292,
                        "feature": "ModelID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "Medium",
                        "strength": 3432.899846833262,
                        "feature": "ProductSize",
                        "label": "SalePrice",
                    },
                ],
            },
        ],
    }
)


class TestPredictJob(SDKTestcase):
    def setUp(self):
        super(TestPredictJob, self).setUp()
        self.project_id = "556902e8100d2b3728d47551"
        self.model_id = "556902ef100d2b37da13077d"
        self.predict_job_id = 111
        self.post_url = "https://host_name.com/projects/{}/predictions/".format(self.project_id)
        self.get_job_url = "https://host_name.com/projects/{}/predictJobs/{}/".format(
            self.project_id, self.predict_job_id
        )
        self.get_predictions_url = "https://host_name.com/projects/{}/predictions/123/".format(
            self.project_id
        )
        self.predict_job_data = {
            "status": "queue",
            "id": self.predict_job_id,
            "projectId": self.project_id,
            "modelId": self.model_id,
            "isBlocked": False,
        }

        self.predict_job_with_message = {
            "status": "error",
            "id": self.predict_job_id,
            "projectId": self.project_id,
            "modelId": self.model_id,
            "isBlocked": False,
            "message": "Clearly this is user error :)",
        }

        self.mock_generic_job_data = {
            "status": "inprogress",
            "url": "https://host_name.com/projects/{}/predictJobs/1/".format(self.project_id),
            "id": "1",
            "jobType": "predict",
            "isBlocked": False,
            "projectId": self.project_id,
        }

        self.mock_model_job_data = {
            "status": "inprogress",
            "url": "https://host_name.com/projects/{}/modelJobs/2/".format(self.project_id),
            "id": "2",
            "jobType": "model",
            "isBlocked": False,
            "projectId": self.project_id,
        }

    def test_instantiate_predict_job(self):
        job = PredictJob(self.predict_job_data)
        print(job)  # checks that method relevant to print works on this (failed in one release)
        self.assertEqual(job.id, self.predict_job_id)
        self.assertEqual(job.status, "queue")
        self.assertIsInstance(job.project, Project)
        self.assertEqual(job.project.id, "556902e8100d2b3728d47551")

        with self.assertRaises(ValueError):
            PredictJob("qid")

    def test_future_proof(self):
        PredictJob(dict(self.predict_job_data, future="new"))

    @responses.activate
    def test_from_job(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/predictJobs/1/".format(self.project_id),
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type="application/json",
        )
        generic_job = Job(self.mock_generic_job_data)
        predict_job = PredictJob.from_job(generic_job)

        self.assertEqual(predict_job.status, "error")
        self.assertEqual(predict_job.job_type, "predict")
        self.assertEqual(predict_job.project, Project(self.project_id))
        self.assertEqual(predict_job.model.id, self.model_id)

    def test_from_job_wrong_type(self):
        generic_job = Job(self.mock_model_job_data)
        with self.assertRaises(ValueError):
            PredictJob.from_job(generic_job)

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_create_non_ascii_filename(self):
        path = fixture_file_path(u"日本/データ.csv")
        model = mock.Mock()
        model.id = self.model_id
        model.project.id = self.project_id
        responses.add(
            responses.POST,
            self.post_url,
            status=202,
            body="",
            adding_headers={"Location": "https://host_name.com" + self.get_job_url},
        )
        predict_job_id = PredictJob.create(model, sourcedata=path)
        assert predict_job_id == str(self.predict_job_id)

    @pytest.mark.usefixtures("known_warning")
    def test_create_bad_sourcedata(self):
        with pytest.raises(errors.InputNotUnderstoodError) as exc_info:
            PredictJob.create(mock.Mock(), sourcedata=123)
        assert_raised_regex(exc_info, "sourcedata parameter not understood.")

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_create_success(self):
        model = mock.Mock()
        model.id = self.model_id
        model.project.id = self.project_id
        responses.add(
            responses.POST,
            self.post_url,
            status=202,
            body="",
            adding_headers={"Location": "https://host_name.com" + self.get_job_url},
        )
        predict_job_id = PredictJob.create(model, b"one,two\n3,4")
        self.assertEqual(predict_job_id, str(self.predict_job_id))

    @responses.activate
    def test_cancel(self):
        predict_job = PredictJob(self.predict_job_data)
        responses.add(
            responses.DELETE, self.get_job_url, status=204, body="", content_type="application/json"
        )
        predict_job.cancel()
        self.assertEqual(responses.calls[0].request.method, "DELETE")

    @responses.activate
    def test_get_unfinished_predict_job_with_message(self):
        """A new attribute of PredictJob beginning in DataRobot API v2.1"""
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type="application/json",
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, "error")
        self.assertEqual(predict_job.message, "Clearly this is user error :)")

    @responses.activate
    def test_get_unfinished_predict_job(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type="application/json",
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, "queue")
        self.assertEqual(predict_job.message, "")

    @responses.activate
    def test_get_finished_predict_job(self):
        responses.add(
            responses.GET, self.get_job_url, status=303, body="",
        )
        with self.assertRaises(errors.PendingJobFinished):
            PredictJob.get(self.project_id, self.predict_job_id)

    @responses.activate
    def test_get_predict_job_async_failure(self):
        responses.add(
            responses.GET, self.get_job_url, status=202, body="",
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code 202")

    @responses.activate
    def test_get_unfinished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type="application/json",
        )
        with pytest.raises(errors.JobNotFinished) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Pending job status is queue")

    @responses.activate
    def test_get_predictions_async_failure(self):
        responses.add(
            responses.GET, self.get_job_url, status=202, body="",
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code 202")

    @responses.activate
    def test_get_finished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body="",
            adding_headers={"Location": self.get_predictions_url},
        )
        responses.add(
            responses.GET,
            self.get_predictions_url,
            status=200,
            body=SAMPLE_BINARY_PREDICTION_RETURN,
            content_type="application/json",
        )
        predictions = PredictJob.get_predictions(self.project_id, self.predict_job_id)
        self.assertIsInstance(predictions, pd.DataFrame)
        self.assertEqual(predictions.shape, (3, 5))
        self.assertEqual(
            sorted(predictions.columns),
            sorted(["positive_probability", "prediction", "row_id", "class_0.0", "class_1.0"]),
        )


class TestWaitForAsyncPredictions(SDKTestcase):
    def setUp(self):
        super(TestWaitForAsyncPredictions, self).setUp()
        self.pid = "p-id"
        self.predict_job_id = "5"
        self.get_job_url = "https://host_name.com/projects/{}/predictJobs/{}/".format(
            self.pid, self.predict_job_id,
        )

    @responses.activate
    def test_timeout(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps({"isBlocked": False}),
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncTimeoutError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id, max_wait=1)
        assert_raised_regex(exc_info, "Predictions generation timed out in")

    @responses.activate
    @mock.patch("datarobot.models.predict_job.PredictJob.get_predictions")
    def test_success(self, get_predictions):
        responses.add(
            responses.GET, self.get_job_url, status=303, body="", content_type="application/json"
        )
        wait_for_async_predictions(self.pid, self.predict_job_id)
        self.assertEqual(get_predictions.call_count, 1)

    @responses.activate
    @mock.patch("datarobot.models.predict_job.PredictJob.get_predictions")
    def test_error(self, get_predictions):
        data = {"status": "error", "isBlocked": False}
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(data),
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncPredictionsGenerationError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id)
        assert_raised_regex(exc_info, "Predictions generation unsuccessful")


class TestRawPredictionResponseToDataframe(unittest.TestCase):
    def test_parse_regression_predictions(self):
        data = json.loads(SAMPLE_REGRESSION_PREDICTION_RETURN)

        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 2))
        self.assertEqual(
            sorted(frame.columns), [u"prediction", u"row_id"],
        )

    def test_parse_classification_predictions(self):
        data = json.loads(SAMPLE_BINARY_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"positive_probability", u"prediction", u"row_id", u"0.0", u"1.0"]),
        )

    def test_parse_classification_predictions_with_floats(self):
        data = json.loads(SAMPLE_BINARY_FLOAT_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "class_")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"positive_probability", u"prediction", u"row_id", u"class_0.0", u"class_1.0"]),
        )

    def test_parse_multiclass_classification_predictions(self):
        data = json.loads(SAMPLE_MULTICLASS_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"prediction", u"row_id", u"class_a", u"class_b", u"class_c"]),
        )

    def test_parse_multiclass_classification_predictions_unicode(self):
        data = json.loads(UNICODE_SAMPLE_MULTICLASS_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"prediction", u"row_id", u"class_赤木", u"class_エピソード", u"class_外"]),
        )

    def test_parse_regression_predictions_with_shap_explanations(self):
        data = json.loads(SAMPLE_REGRESSION_PREDICTION_WITH_SHAP_EXPL_RETURN)
        frame = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame.shape == (2, 13)
        assert sorted(frame.columns) == sorted(
            [
                "prediction",
                "row_id",
                "Explanation_1_feature_name",
                "Explanation_1_feature_value",
                "Explanation_1_strength",
                "Explanation_2_feature_name",
                "Explanation_2_feature_value",
                "Explanation_2_strength",
                "Explanation_3_feature_name",
                "Explanation_3_feature_value",
                "Explanation_3_strength",
                "shap_remaining_total",
                "shap_base_value",
            ]
        )

        # assert scalar shapBaseValue doesn't cause an error
        data["shapBaseValue"] = data["shapBaseValue"][0]
        frame_2 = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame.equals(frame_2)

        # when shapBaseValue has multiple items the first one is used and no error is raised
        data["shapBaseValue"] = [9.12, 42.1, -27.1]
        frame_3 = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame_3["shap_base_value"].tolist() == [9.12, 9.12]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import re
import sys

from setuptools.command.test import test as TestCommand

DESCRIPTION_TEMPLATE = """
About {package_name}
============================
.. image:: https://img.shields.io/pypi/v/{package_name}.svg
   :target: {pypi_url_target}
.. image:: https://img.shields.io/pypi/pyversions/{package_name}.svg
.. image:: https://img.shields.io/pypi/status/{package_name}.svg

DataRobot is a client library for working with the `DataRobot`_ platform API. {extra_desc}

This package is released under the terms of the DataRobot Tool and Utility Agreement, which can be
 found on our `Legal`_ page, along with our privacy policy and more.

Installation
=========================
Python {python_versions} are supported.
You must have a datarobot account.

::

   $ pip install {pip_package_name}

Usage
=========================
The library will look for a config file `~/.config/datarobot/drconfig.yaml` by default.
This is an example of what that config file should look like.

::

   token: your_token
   endpoint: https://app.datarobot.com/api/v2

Alternatively a global client can be set in the code.

::

   import datarobot as dr
   dr.Client(token='your_token', endpoint='https://app.datarobot.com/api/v2')

Alternatively environment variables can be used.

::

   export DATAROBOT_API_TOKEN='your_token'
   export DATAROBOT_ENDPOINT='https://app.datarobot.com/api/v2'

See `documentation`_ for example usage after configuring.

Tests
=========================
::

   $ py.test

.. _datarobot: http://datarobot.com
.. _documentation: {docs_link}
.. _legal: https://www.datarobot.com/legal/
"""


DEFAULT_CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with open("datarobot/_version.py") as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

lint_require = [
    'black==19.10b0; python_version >= "3.6"',
    'isort==5.8; python_version >= "3.6"',
]

tests_require = [
    "mock==3.0.5",
    "pytest>=4.6,<5",  # 4.6 is last release supporting python2
    "pytest-cov",
    "responses>=0.9,<0.10",  # 0.10 changes the interface, will require test changes
]

_decorator_constraints = [
    'decorator<5; python_version=="2.7"',
]

dev_require = (
    tests_require
    + lint_require
    + _decorator_constraints
    + [
        # One of the packages in this list is depending on a broken version of docutils, causing
        # test-docs to fail, so we pin it here, despite this package doesn't really need it as such.
        #  See more: https://github.com/NVIDIAGameWorks/kaolin/issues/76
        "docutils==0.15.2",
        "flake8>=2.5.2,<3",
        "Sphinx==1.8.3",
        "sphinx_rtd_theme==0.1.9",
        "nbsphinx>=0.2.9,<1",
        "mistune==0.8.4",
        "nbconvert==5.3.1",
        "numpydoc>=0.6.0",
        "tox",
        "jupyter_contrib_nbextensions",
        "tornado<6.0",
        "jsonschema<=3.2.0",
    ]
)

example_require = _decorator_constraints + [
    "jupyter<=5.0",
    "fredapi==0.4.0",
    "matplotlib>=2.1.0",
    "seaborn<=0.8",
    "scikit-learn<=0.18.2",
    "wordcloud<=1.3.1",
    "colour<=0.1.4",
]

release_require = ["zest.releaser[recommended]==6.22.0"]

# The None-valued kwargs should be updated by the caller
common_setup_kwargs = dict(
    name=None,
    version=None,
    description="This client library is designed to support the DataRobot API.",
    author="datarobot",
    author_email="support@datarobot.com",
    maintainer="datarobot",
    maintainer_email="info@datarobot.com",
    url="http://datarobot.com",
    license="DataRobot Tool and Utility Agreement",
    packages=None,
    long_description=None,
    classifiers=None,
    install_requires=[
        "contextlib2>=0.5.5",
        "pandas>=0.15,<1.3.1",
        "pyyaml>=3.11",
        "requests>=2.21",
        "requests_toolbelt>=0.6",
        "trafaret>=0.7,<2.2,!=1.1.0",
        "urllib3>=1.23",
        "attrs>=19.1.0,<21.0",
    ],
    tests_require=tests_require,
    extras_require={
        "dev": dev_require,
        "examples": example_require,
        "release": release_require,
        "lint": lint_require,
    },
    cmdclass={"test": PyTest},
)



# -*- coding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
#
# DataRobot Python API documentation build configuration file, created by
# sphinx-quickstart on Mon Mar 16 10:50:19 2015.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import shutil
import sys
from zipfile import ZipFile

sys.path.insert(0, os.path.abspath(".."))
from datarobot import __version__ as version  # noqa

# Make sure magic 'tags' field is defined.
# (Not actually useful but makes Cython and pylint errors go away.)
if "tags" not in locals():
    tags = None
if "b" in version:
    # mainline quarterly releases never have the beta "b" in the version, only internal & weekly do
    tags.add("include_experimental_docs")

# Copy the changelog to the docs directory so sphinx will include it
root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
changelog_path = os.path.join(root_dir, "CHANGES.rst")
dest_path = os.path.join(root_dir, "docs", "CHANGES.rst")
assert os.path.isfile(changelog_path)
shutil.copyfile(changelog_path, dest_path)

# Create a zip file of the example resources (csv, py, ipynb)
# so that it can be included it as a download within the HTML docs
files_to_zip = []
for root, _, files in os.walk("examples"):
    basename = os.path.basename(root)
    # skip "hidden" directories like notebook autosaves and _build
    if not basename.startswith("_") and not basename.startswith("."):
        for file in files:
            name, ext = os.path.splitext(file)
            ext_lower = ext.lower()
            if ext_lower in [".csv", ".xls", ".xlsx", ".ipynb", ".py"]:
                files_to_zip.append(os.path.join(root, file))
with ZipFile("datarobot_examples.zip", "w") as examples_zip:
    for file in files_to_zip:
        examples_zip.write(file)

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "numpydoc",
    "nbsphinx",
]

# make warnings go away- as advised by
# https://github.com/phn/pytpm/issues/3#issuecomment-12133978
numpydoc_show_class_members = False

nbsphinx_execute = "never"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# Showing automatically documented members in same order as they are in source
autodoc_member_order = "bysource"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# Thes are both the full X.Y.Z version, including alpha/beta/rc tags.
# We think Sphinx probably wants both to be defined.
release = version

project = u"DataRobot Python Client"
copyright = u"2019, DataRobot, Inc."


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "**/.ipynb_checkpoints"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "DataRobotPythonDoc"


# -- Options for LaTeX output --------------------------------------------------

# latex_elements = {
#     The paper size ('letterpaper' or 'a4paper').
#     'papersize': 'letterpaper',

#     The font size ('10pt', '11pt' or '12pt').
#     'pointsize': '10pt',

#     Additional stuff for the LaTeX preamble.
#     'preamble': '',
# }

fh = open("latex_preamble.tex", "r+")
PREAMBLE = fh.read()
fh.close()
latex_elements = {
    "inputenc": "",
    "utf8extra": "",
    "preamble": PREAMBLE,
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "DataRobotPython_{}_Docs.tex".format(version),
        u"DataRobot Python API Documentation",
        "DataRobot, Inc.",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "datarobotpython", u"DataRobot Python API Documentation", [u"DataRobot, Inc."], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "DataRobotPython",
        u"DataRobot Python API Documentation",
        "DataRobot, Inc.",
        "DataRobotPython",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


def setup(app):
    app.add_css_file("css/tight_table.css")



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
import responses

from datarobot import errors


@responses.activate
@pytest.mark.usefixtures("feature_impact_job_creation_response", "feature_impact_running_response")
def test_get_feature_impact_job_result_not_finished(one_model):
    feature_impact_job = one_model.request_feature_impact()
    with pytest.raises(errors.JobNotFinished):
        feature_impact_job.get_result()


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_job_creation_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_get_feature_impact_job_result_finished(feature_impact_server_data, one_model):
    feature_impact_job = one_model.request_feature_impact(
        row_count=feature_impact_server_data["rowCount"]
    )
    feature_impact = feature_impact_job.get_result()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_job_creation_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_get_feature_impact_job_result_finished_with_metadata(
    feature_impact_server_data, feature_impact_server_data_filtered, one_model
):
    feature_impact_job = one_model.request_feature_impact(
        row_count=feature_impact_server_data["rowCount"], with_metadata=True
    )
    feature_impact = feature_impact_job.get_result()
    # Note: we are not exposing the rowCount via an existing method yet and this is just to ensure
    # it does not break with an extended response format.
    assert feature_impact == feature_impact_server_data_filtered


@responses.activate
@pytest.mark.usefixtures("feature_impact_job_creation_response", "feature_impact_running_response")
def test_wait_for_feature_impact_never_finished(one_model, mock_async_time):
    mock_async_time.time.side_effect = (0, 5)
    feature_impact_job = one_model.request_feature_impact()
    with pytest.raises(errors.AsyncTimeoutError):
        feature_impact_job.get_result_when_complete(max_wait=1)


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_previously_ran_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_get_or_request_feature_impact_previously_requested(one_model, feature_impact_server_data):
    feature_impact = one_model.get_or_request_feature_impact()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_previously_ran_response__no_job_id", "feature_impact_response",
)
def test_get_or_request_feature_impact_previously_requested_but_no_job_id(
    one_model, feature_impact_server_data
):
    feature_impact = one_model.get_or_request_feature_impact()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_previously_ran_response", "feature_impact_running_response",
)
def test_get_or_request_feature_impact_currently_running_waits(one_model, mock_async_time):
    mock_async_time.time.side_effect = (0, 5)
    with pytest.raises(errors.AsyncTimeoutError):
        one_model.get_or_request_feature_impact(max_wait=1)


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_job_creation_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_get_or_request_feature_impact(one_model, feature_impact_server_data):
    feature_impact = one_model.get_or_request_feature_impact()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures(
    "feature_impact_job_creation_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_get_or_request_feature_impact_with_metadata(
    one_model, feature_impact_server_data_filtered
):
    feature_impact = one_model.get_or_request_feature_impact(with_metadata=True)
    assert feature_impact == feature_impact_server_data_filtered


@responses.activate
@pytest.mark.usefixtures(
    "client",
    "feature_impact_job_creation_response",
    "feature_impact_completed_response",
    "feature_impact_response",
)
def test_wait_for_feature_impact_finished(one_model, feature_impact_server_data):
    # Would like to include a mock response for not done yet, but that didn't work.
    # Looks like `responses` might do that when `assert_all_requests_are_fired` is true.
    feature_impact_job = one_model.request_feature_impact()
    feature_impact = feature_impact_job.get_result_when_complete(max_wait=0.5)
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures("feature_impact_response")
def test_get_feature_impact_assumed_complete(one_model, feature_impact_server_data):
    feature_impact = one_model.get_feature_impact()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures("feature_impact_response")
def test_get_feature_impact_assumed_complete_with_metadata(
    one_model, feature_impact_server_data_filtered
):
    feature_impact = one_model.get_feature_impact(with_metadata=True)
    assert feature_impact == feature_impact_server_data_filtered


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_feature_impact_no_redundancy(
    one_model, feature_impact_server_data, feature_impact_url
):
    feature_impact_server_data["ranRedundancyDetection"] = False
    responses.add(
        responses.GET,
        feature_impact_url,
        status=200,
        content_type="application/json",
        json=feature_impact_server_data,
    )
    feature_impact = one_model.get_feature_impact()
    assert feature_impact == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.usefixtures("multiclass_feature_impact_response")
def test_get_multiclass_feature_impact(one_model, multiclass_feature_impact_server_data):
    feature_impact = one_model.get_multiclass_feature_impact()
    assert feature_impact == multiclass_feature_impact_server_data["classFeatureImpacts"]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

import pytest
import responses

from datarobot import DataSource, DataSourceParameters, SharingAccess
from datarobot.utils import parse_time
from tests.utils import request_body_to_json


@responses.activate
def test_list(data_sources_endpoint, data_sources_list_server_resp):
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)

    data_sources = DataSource.list()

    for data_source, server_payload in zip(data_sources, data_sources_list_server_resp["data"]):
        assert data_source.id == server_payload["id"]
        assert data_source.type == server_payload["type"]
        assert data_source.canonical_name == server_payload["canonicalName"]
        assert data_source.creator == server_payload["creator"]
        assert data_source.updated == parse_time(server_payload["updated"])
        assert data_source.params.data_store_id == server_payload["params"]["dataStoreId"]
        # assert params are the same
        params = data_source.params
        srv_params = server_payload["params"]
        assert params.table == (srv_params["table"] if "table" in srv_params else None)
        assert params.schema == (srv_params["schema"] if "schema" in srv_params else None)
        assert params.partition_column == (
            srv_params["partitionColumn"] if "partitionColumn" in srv_params else None
        )
        assert params.query == (srv_params["query"] if "query" in srv_params else None)
        assert params.fetch_size == (srv_params["fetchSize"] if "fetchSize" in srv_params else None)


@responses.activate
def test_get(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)

    assert data_source.id == diagnostics_data_source_server_resp["id"]
    assert data_source.type == diagnostics_data_source_server_resp["type"]
    assert data_source.canonical_name == diagnostics_data_source_server_resp["canonicalName"]
    assert data_source.creator == diagnostics_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(diagnostics_data_source_server_resp["updated"])
    assert (
        data_source.params.data_store_id
        == diagnostics_data_source_server_resp["params"]["dataStoreId"]
    )
    # assert params are the same
    params = data_source.params
    srv_params = diagnostics_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.query == srv_params["query"]


@responses.activate
def test_future_proof(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp["id"]
    aug_resp = dict(diagnostics_data_source_server_resp, newest="newest")
    aug_resp["params"]["newKey"] = "hihihi"
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)

    assert data_source.id == data_source_id


@responses.activate
def test_create_from_query(data_sources_endpoint, diagnostics_data_source_server_resp):
    responses.add(responses.POST, data_sources_endpoint, json=diagnostics_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=diagnostics_data_source_server_resp["params"]["dataStoreId"],
        query=diagnostics_data_source_server_resp["params"]["query"],
    )

    data_source = DataSource.create(
        data_source_type=diagnostics_data_source_server_resp["type"],
        canonical_name=diagnostics_data_source_server_resp["canonicalName"],
        params=params,
    )

    assert data_source.id == diagnostics_data_source_server_resp["id"]
    assert data_source.type == diagnostics_data_source_server_resp["type"]
    assert data_source.canonical_name == diagnostics_data_source_server_resp["canonicalName"]
    assert data_source.creator == diagnostics_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(diagnostics_data_source_server_resp["updated"])
    # assert params are the same
    params = data_source.params
    srv_params = diagnostics_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.query == srv_params["query"]
    assert params.partition_column is None
    assert params.table is None
    assert params.schema is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table(
    data_sources_endpoint, airlines10mb_1000_records_data_source_server_resp
):
    del airlines10mb_1000_records_data_source_server_resp["params"]["fetchSize"]
    responses.add(
        responses.POST,
        data_sources_endpoint,
        json=airlines10mb_1000_records_data_source_server_resp,
    )
    params = DataSourceParameters(
        data_store_id=airlines10mb_1000_records_data_source_server_resp["params"]["dataStoreId"],
        table=airlines10mb_1000_records_data_source_server_resp["params"]["table"],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_1000_records_data_source_server_resp["type"],
        canonical_name=airlines10mb_1000_records_data_source_server_resp["canonicalName"],
        params=params,
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp["id"]
    assert data_source.type == airlines10mb_1000_records_data_source_server_resp["type"]
    assert (
        data_source.canonical_name
        == airlines10mb_1000_records_data_source_server_resp["canonicalName"]
    )
    assert data_source.creator == airlines10mb_1000_records_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(
        airlines10mb_1000_records_data_source_server_resp["updated"]
    )
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_1000_records_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.table == srv_params["table"]
    assert params.partition_column is None
    assert params.query is None
    assert params.schema is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table_limited_with_fetch_size(
    data_sources_endpoint, airlines10mb_1000_records_data_source_server_resp
):
    responses.add(
        responses.POST,
        data_sources_endpoint,
        json=airlines10mb_1000_records_data_source_server_resp,
    )
    params = DataSourceParameters(
        data_store_id=airlines10mb_1000_records_data_source_server_resp["params"]["dataStoreId"],
        table=airlines10mb_1000_records_data_source_server_resp["params"]["table"],
        fetch_size=airlines10mb_1000_records_data_source_server_resp["params"]["fetchSize"],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_1000_records_data_source_server_resp["type"],
        canonical_name=airlines10mb_1000_records_data_source_server_resp["canonicalName"],
        params=params,
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp["id"]
    assert data_source.type == airlines10mb_1000_records_data_source_server_resp["type"]
    assert (
        data_source.canonical_name
        == airlines10mb_1000_records_data_source_server_resp["canonicalName"]
    )
    assert data_source.creator == airlines10mb_1000_records_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(
        airlines10mb_1000_records_data_source_server_resp["updated"]
    )
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_1000_records_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.table == srv_params["table"]
    assert params.fetch_size == srv_params["fetchSize"]
    assert params.partition_column is None
    assert params.query is None
    assert params.schema is None


@responses.activate
def test_create_from_table_specific_schema(
    data_sources_endpoint, airlines10mb_data_source_server_resp
):
    responses.add(responses.POST, data_sources_endpoint, json=airlines10mb_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_data_source_server_resp["params"]["dataStoreId"],
        table=airlines10mb_data_source_server_resp["params"]["table"],
        schema=airlines10mb_data_source_server_resp["params"]["schema"],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_data_source_server_resp["type"],
        canonical_name=airlines10mb_data_source_server_resp["canonicalName"],
        params=params,
    )

    assert data_source.id == airlines10mb_data_source_server_resp["id"]
    assert data_source.type == airlines10mb_data_source_server_resp["type"]
    assert data_source.canonical_name == airlines10mb_data_source_server_resp["canonicalName"]
    assert data_source.creator == airlines10mb_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(airlines10mb_data_source_server_resp["updated"])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.table == srv_params["table"]
    assert params.schema == srv_params["schema"]
    assert params.partition_column is None
    assert params.query is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table_with_partition_column(
    data_sources_endpoint, airlines10mb_data_source_server_resp
):
    del airlines10mb_data_source_server_resp["params"]["schema"]
    airlines10mb_data_source_server_resp["params"]["partitionColumn"] = "part_column"
    responses.add(responses.POST, data_sources_endpoint, json=airlines10mb_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_data_source_server_resp["params"]["dataStoreId"],
        table=airlines10mb_data_source_server_resp["params"]["table"],
        partition_column=airlines10mb_data_source_server_resp["params"]["partitionColumn"],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_data_source_server_resp["type"],
        canonical_name=airlines10mb_data_source_server_resp["canonicalName"],
        params=params,
    )

    assert data_source.id == airlines10mb_data_source_server_resp["id"]
    assert data_source.type == airlines10mb_data_source_server_resp["type"]
    assert data_source.canonical_name == airlines10mb_data_source_server_resp["canonicalName"]
    assert data_source.creator == airlines10mb_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(airlines10mb_data_source_server_resp["updated"])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_data_source_server_resp["params"]
    assert params.data_store_id == srv_params["dataStoreId"]
    assert params.table == srv_params["table"]
    assert params.partition_column == srv_params["partitionColumn"]
    assert params.schema is None
    assert params.query is None
    assert params.fetch_size is None


@responses.activate
def test_update_to_table(
    data_sources_endpoint,
    diagnostics_data_source_server_resp,
    airlines10mb_1000_records_data_source_server_resp,
):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )
    updated_params = DataSourceParameters(
        table=airlines10mb_1000_records_data_source_server_resp["params"]["table"],
        fetch_size=airlines10mb_1000_records_data_source_server_resp["params"]["fetchSize"],
    )
    updated_data_source_server_resp = dict(
        diagnostics_data_source_server_resp,
        canonicalName="updated_canonical_name",
        params=updated_params.collect_payload(),
    )
    updated_data_source_server_resp["params"]["dataStoreId"] = diagnostics_data_source_server_resp[
        "params"
    ]["dataStoreId"]
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp,
    )
    data_source = DataSource.get(data_source_id)

    data_source.update(
        canonical_name=updated_data_source_server_resp["canonicalName"], params=updated_params
    )

    assert data_source.id == updated_data_source_server_resp["id"]
    assert data_source.type == updated_data_source_server_resp["type"]
    assert data_source.canonical_name == updated_data_source_server_resp["canonicalName"]
    assert data_source.creator == updated_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(updated_data_source_server_resp["updated"])
    assert data_source.params == updated_params


@responses.activate
def test_update_fetch_size(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    fetch_size = 12345
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    airlines10mb_data_source_server_resp["params"]["fetchSize"] = fetch_size
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    assert data_source.params.fetch_size is None

    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs["fetch_size"] = fetch_size
    data_source.update(params=DataSourceParameters(**new_perams_kwargs))
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.params.fetch_size == fetch_size


@responses.activate
def test_update_schema(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    updated_schema = "other_schema"
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    airlines10mb_data_source_server_resp["params"]["schema"] = updated_schema
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs["schema"] = updated_schema
    data_source.update(params=DataSourceParameters(**new_perams_kwargs))
    new_data_source = DataSource.get(data_source_id)
    assert new_data_source.params.schema == updated_schema


@responses.activate
def test_update_partition_column(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    partition_column = "my_partition_column"
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    airlines10mb_data_source_server_resp["params"]["partition_column"] = partition_column
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs["partition_column"] = partition_column
    data_source.update(params=DataSourceParameters(**new_perams_kwargs))
    new_data_source = DataSource.get(data_source_id)
    assert new_data_source.params.partition_column == partition_column


@responses.activate
def test_update_to_query(
    data_sources_endpoint,
    diagnostics_data_source_server_resp,
    airlines10mb_1000_records_data_source_server_resp,
):
    data_source_id = airlines10mb_1000_records_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=airlines10mb_1000_records_data_source_server_resp,
    )
    updated_params = DataSourceParameters(
        query=diagnostics_data_source_server_resp["params"]["query"],
    )
    updated_data_source_server_resp = dict(
        airlines10mb_1000_records_data_source_server_resp,
        canonicalName="updated_canonical_name",
        params=updated_params.collect_payload(),
    )
    updated_data_source_server_resp["params"][
        "dataStoreId"
    ] = airlines10mb_1000_records_data_source_server_resp["params"]["dataStoreId"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp,
    )
    data_source = DataSource.get(data_source_id)

    data_source.update(
        canonical_name=updated_data_source_server_resp["canonicalName"], params=updated_params
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp["id"]
    assert data_source.type == updated_data_source_server_resp["type"]
    assert data_source.canonical_name == updated_data_source_server_resp["canonicalName"]
    assert data_source.creator == updated_data_source_server_resp["creator"]
    assert data_source.updated == parse_time(updated_data_source_server_resp["updated"])
    assert data_source.params == updated_params


@responses.activate
def test_update_unable_to_change_type(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    data_source._type = "new_type"
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.type == diagnostics_data_source_server_resp["type"]


@responses.activate
def test_update_unable_to_change_creator(
    data_sources_endpoint, diagnostics_data_source_server_resp
):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    data_source._created = "other_user"
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.creator == diagnostics_data_source_server_resp["creator"]


@responses.activate
def test_update_unable_to_change_updated(
    data_sources_endpoint, diagnostics_data_source_server_resp
):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )

    data_source = DataSource.get(data_source_id)
    data_source._updated = datetime.now()
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.updated == parse_time(diagnostics_data_source_server_resp["updated"])


@responses.activate
def test_delete(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp,
    )
    delete_url = "{}{}/".format(data_sources_endpoint, data_source_id)
    responses.add(responses.DELETE, delete_url)

    data_source = DataSource.get(data_source_id)
    data_source.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url


@responses.activate
@pytest.mark.parametrize("extra_keys", [{}, {"newKey": "blah"}])
def test_get_access_list(data_sources_endpoint, diagnostics_data_source_server_resp, extra_keys):
    data_source = DataSource.from_server_data(diagnostics_data_source_server_resp)
    access_list = {
        "data": [
            {
                "username": "me@datarobot.com",
                "userId": "1234deadbeeffeeddead4321",
                "role": "OWNER",
                "canShare": True,
            }
        ],
        "count": 1,
        "previous": None,
        "next": None,
    }
    access_record = access_list["data"][0]
    access_record.update(extra_keys)
    responses.add(
        responses.GET,
        "{}{}/accessControl/".format(data_sources_endpoint, data_source.id),
        json=access_list,
    )

    response = data_source.get_access_list()
    assert len(response) == 1

    share_info = response[0]
    assert share_info.username == access_record["username"]
    assert share_info.user_id == access_record["userId"]
    assert share_info.role == access_record["role"]
    assert share_info.can_share == access_record["canShare"]


@responses.activate
def test_share(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source = DataSource.from_server_data(diagnostics_data_source_server_resp)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_sources_endpoint, data_source.id),
        status=204,
    )

    data_source.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": "EDITOR"}]}


@responses.activate
def test_remove(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source = DataSource.from_server_data(diagnostics_data_source_server_resp)
    request = SharingAccess("me@datarobot.com", None)
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_sources_endpoint, data_source.id),
        status=204,
    )

    data_source.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": None}]}


@responses.activate
def test_share_with_grant(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source = DataSource.from_server_data(diagnostics_data_source_server_resp)
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_sources_endpoint, data_source.id),
        status=204,
    )

    data_source.share(
        [
            SharingAccess("me@datarobot.com", "EDITOR", can_share=True),
            SharingAccess("other@datarobot.com", "CONSUMER", can_share=False),
        ]
    )

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "data": [
            {"username": "me@datarobot.com", "role": "EDITOR", "canShare": True},
            {"username": "other@datarobot.com", "role": "CONSUMER", "canShare": False},
        ]
    }



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import Project


@pytest.fixture
def project(project_with_target_data):
    return Project.from_data(project_with_target_data)


@responses.activate
def test_get_feature_association_details(
    project,
    project_id,
    feature_association_matrix_details_url,
    feature_association_matrix_details_data,
):
    feature1 = "area1"
    feature2 = "area2"
    responses.add(
        responses.GET,
        feature_association_matrix_details_url.format(project_id, feature1, feature2),
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix_details_data),
    )
    feature_values = project.get_association_matrix_details(feature1="area1", feature2="area2")

    assert set(feature_values.keys()) == {"chart_type", "values", "features", "types"}
    assert len(feature_values["values"]) == 7
    assert feature_values["features"] == ["area1", "area2"]


@responses.activate
def test_get_feature_association_featurelists(
    feature_association_featurelists_data, feature_association_featurelists_url, project, project_id
):
    responses.add(
        responses.GET,
        feature_association_featurelists_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_featurelists_data),
    )
    featurelists = project.get_association_featurelists()

    assert set(featurelists.keys()) == {"featurelists"}
    assert len(featurelists["featurelists"]) == 2
    assert featurelists["featurelists"][0]["title"] == "Informative Features"
    assert featurelists["featurelists"][0]["has_fam"] is True


@responses.activate
def test_get_feature_associations(
    project, feature_association_matrix_url, feature_association_matrix
):
    responses.add(
        responses.GET,
        feature_association_matrix_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix),
    )
    feature_associations = project.get_associations(metric=None, assoc_type=None)
    assert feature_associations
    assert set(feature_associations.keys()) == {"strengths", "features"}



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import Int, String

from ..api_object import APIObject
from .images import Image

__all__ = ["ImageAugmentationOptions", "ImageAugmentationList", "ImageAugmentationSample"]


class ImageAugmentationOptions(APIObject):
    """A List of all supported Image Augmentation Transformations for a project.
    Includes additional information about minimum, maximum, and default values
    for a transformation.

    Attributes
    ----------
    name: string
        The name of the augmentation list
    project_id: string
        The project containing the image data to be augmented
    min_transformation_probability: float
        The minimum allowed value for transformation probability.
    current_transformation_probability: float
        Default setting for probability that each transformation will be applied to an image.
    max_transformation_probability: float
        The maximum allowed value for transformation probability.
    min_number_of_new_images: int
         The minimum allowed number of new rows to add for each existing row
    current_number_of_new_images: int
         The default number of new rows to add for each existing row
    max_number_of_new_images: int
         The maximum allowed number of new rows to add for each existing row
    transformations: array
        List of transformations to possibly apply to each image
    """

    _get_path = "imageAugmentationOptions/{pid}"
    _converter = t.Dict(
        {
            t.Key("project_id"): String(),
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("min_transformation_probability"): t.Float(),
            t.Key("max_transformation_probability"): t.Float(),
            t.Key("current_transformation_probability"): t.Float(),
            t.Key("min_number_of_new_images"): Int(),
            t.Key("max_number_of_new_images"): Int(),
            t.Key("current_number_of_new_images"): Int(),
            t.Key("transformations", optional=True): t.List(t.Dict().allow_extra("*")),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id,
        name,
        project_id,
        min_transformation_probability,
        current_transformation_probability,
        max_transformation_probability,
        min_number_of_new_images,
        current_number_of_new_images,
        max_number_of_new_images,
        transformations=None,
    ):
        self.id = id
        self.name = name
        self.project_id = project_id
        self.min_transformation_probability = min_transformation_probability
        self.current_transformation_probability = current_transformation_probability
        self.max_transformation_probability = max_transformation_probability
        self.min_number_of_new_images = min_number_of_new_images
        self.current_number_of_new_images = current_number_of_new_images
        self.max_number_of_new_images = max_number_of_new_images
        self.transformations = transformations

    def __repr__(self):
        return (
            "datarobot.models.visualai.ImageAugmentationOptions("
            "id={0.id}, "
            "name={0.name}, "
            "project_id={0.project_id}, "
            "min_transformation_probability={0.min_transformation_probability}, "
            "max_transformation_probability={0.max_transformation_probability}, "
            "current_transformation_probability={0.current_transformation_probability}, "
            "min_number_of_new_images={0.min_number_of_new_images})"
            "max_number_of_new_images={0.max_number_of_new_images})"
            "current_number_of_new_images={0.current_number_of_new_images})"
        ).format(self)

    @classmethod
    def get(cls, project_id):
        """
        Returns a list of all supported transformations for the given
        project

        :param project_id: sting
            The id of the project for which to return the list of supported transformations.

        :return:
          ImageAugmentationOptions
           A list containing all the supported transformations for the project.
        """
        path = cls._get_path.format(pid=project_id)
        server_data = cls._client.get(path)
        return cls.from_server_data(server_data.json())


class ImageAugmentationList(APIObject):

    """A List of Image Augmentation Transformations

    Attributes
    ----------
    name: string
        The name of the augmentation list
    project_id: string
        The project containing the image data to be augmented
    feature_name: string (optional)
        name of the feature that the augmentation list is associated with
    in_use: boolean
        Whether this is the list that will passed in to every blueprint during blueprint generation
        before autopilot
    initial_list: boolean
        True if this is the list to be used during training to produce augmentations
    transformation_probability: float
        Probability that each transformation will be applied to an image.  Value should be
        between 0.01 - 1.0.
    number_of_new_images: int
         Number of new rows to add for each existing row
    transformations: array
        List of transformations to possibly apply to each image
    """

    _create_path = "imageAugmentationLists/"
    _converter = t.Dict(
        {
            t.Key("project_id"): String(),
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("feature_name", optional=True): String(),
            t.Key("in_use", optional=True): t.Bool(),
            t.Key("initial_list", optional=True): t.Bool(),
            t.Key("transformation_probability", optional=True): t.Float(),
            t.Key("number_of_new_images", optional=True, default=1): Int(),
            t.Key("transformations", optional=True): t.List(t.Dict().allow_extra("*")),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id,
        name,
        project_id,
        feature_name=None,
        in_use=False,
        initial_list=False,
        transformation_probability=0.0,
        number_of_new_images=1,
        transformations=None,
    ):
        self.id = id
        self.name = name
        self.project_id = project_id
        self.feature_name = feature_name
        self.in_use = in_use
        self.initial_list = initial_list
        self.transformation_probability = transformation_probability
        self.number_of_new_images = number_of_new_images
        self.transformations = transformations

    def __repr__(self):
        return (
            "datarobot.models.visualai.ImageAugmentationList("
            "aug_id={0.id}, "
            "name={0.name}, "
            "project_id={0.project_id}, "
            "feature_name={0.feature_name}, "
            "in_use={0.in_use}, "
            "initial_list={0.initial_list}, "
            "transformation_probability={0.transformation_probability}, "
            "number_of_new_images={0.number_of_new_images})"
        ).format(self)

    @classmethod
    def create(
        cls,
        name,
        project_id,
        feature_name=None,
        in_use=False,
        initial_list=False,
        transformation_probability=0.0,
        number_of_new_images=1,
        transformations=None,
    ):
        """
        create a new image augmentation list
        """
        data = {
            "name": name,
            "project_id": project_id,
            "feature_name": feature_name,
            "in_use": in_use,
            "initial_list": initial_list,
            "transformation_probability": transformation_probability,
            "number_of_new_images": number_of_new_images,
            "transformations": transformations,
        }
        server_data = cls._client.post(cls._create_path, data=data)
        list_id = server_data.json()["augmentationListId"]
        return cls.get(list_id)

    @classmethod
    def get(cls, list_id):
        path = cls._create_path + str(list_id) + "/"
        server_data = cls._client.get(path)
        return cls.from_server_data(server_data.json())

    @classmethod
    def delete(cls, list_id):
        path = cls._create_path + str(list_id) + "/"
        cls._client.delete(path)


class ImageAugmentationSample(APIObject):
    """
    A preview of the type of images that augmentations will create during training.

     Attributes
    ----------
    sample_id : ObjectId
        The id of the augmentation sample, used to group related images together
    image_id : ObjectId
        A reference to the Image which can be used to retrieve the image binary
    project_id : ObjectId
        A reference to the project containing the image
    original_image_id : ObjectId
        A reference to the original image that generated this image in the case of an augmented
        image.  If this is None it signifies this is an original image
    height : int
        Image height in pixels
    width : int
        Image width in pixels
    """

    _compute_path = "imageAugmentationSamples/"
    _list_path = "imageAugmentationSamples/{sample_id}/"
    _converter = t.Dict(
        {
            t.Key("image_id"): String(),
            t.Key("project_id"): String(),
            t.Key("height"): Int(),
            t.Key("width"): Int(),
            t.Key("original_image_id", optional=True): String(),
            t.Key("sample_id", optional=True): String(),
        }
    ).ignore_extra("*")

    def __init__(self, image_id, project_id, height, width, original_image_id=None, sample_id=None):
        self.sample_id = sample_id
        self.image_id = image_id
        self.project_id = project_id
        self.original_image_id = original_image_id
        self.height = height
        self.width = width
        self.image = Image(image_id=image_id, project_id=project_id, height=height, width=width)

    def __repr__(self):
        return (
            "datarobot.models.visualai.ImageAugmentationSample("
            "image_id={0.image_id}, "
            "project_id={0.project_id}, "
            "height={0.height}, "
            "width={0.width}, "
            "original_image_id={0.original_image_id})"
            "sample_id={0.sample_id}, "
        ).format(self)

    @classmethod
    def compute(cls, augmentation_list, number_of_rows=5):
        """Start creation of image augmentation samples.

        Parameters
        ----------
        number_of_rows: int
            The number of rows from the original dataset to use as input data for the
            augmentation samples. Defaults to 5.
        augmentation_list: ImageAugmentationList
            An Image Augmentation list that specifies the transformations to apply to each image
            during augmentation.

        Returns
        -------
        str
            URL to check for image augmentation samples generation progress.

        Raises
        ------
        datarobot.errors.ClientError
            Server rejected creation due to client error. Most likely
            cause is bad invalid ``augmentation_list``.
        """
        post_data = augmentation_list.__dict__
        post_data["number_of_rows"] = number_of_rows
        r_data = cls._client.post(cls._compute_path, data=post_data)
        return r_data.headers["Location"]

    @classmethod
    def list(cls, sample_id):
        """Return a list of ImageAugmentationSample objects

         Parameters
         ----------
        sample_id: str
             Unique Id for the set of sample images
        """
        path = cls._list_path.format(sample_id=sample_id)

        result = cls._client.get(path)
        r_data = result.json()
        ret = []
        for sample_data in r_data.get("data", []):
            sample = cls.from_server_data(sample_data)
            sample.sample_id = sample_id
            ret.append(sample)
        return ret



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import itertools
import json
import random

import pytest
from six.moves.urllib.parse import urljoin

from datarobot import TARGET_TYPE


@pytest.fixture
def visualai_url(project_url):
    return urljoin(project_url, "images/")


@pytest.fixture
def visualai_image_url(visualai_url, visualai_image_id):
    return urljoin(visualai_url, visualai_image_id + "/")


@pytest.fixture
def image_augmentations_url(unittest_endpoint):
    return urljoin(unittest_endpoint, "imageAugmentationLists/")


@pytest.fixture
def augmentation_samples_url(unittest_endpoint):
    return urljoin(unittest_endpoint, "imageAugmentationSamples/")


@pytest.fixture
def status_id():
    return "91e49da5-c548-4a97-8a7d-bde7710a7f4f"


@pytest.fixture
def aug_pid():
    return "5e7e562528513130ab237875"


@pytest.fixture
def augmentation_options_url(unittest_endpoint, aug_pid):
    return urljoin(unittest_endpoint, "imageAugmentationOptions/{}".format(aug_pid))


@pytest.fixture
def status_url(unittest_endpoint, status_id):
    return unittest_endpoint + "/status/{}/".format(status_id)


@pytest.fixture
def visualai_image_id():
    return "5e7e562528513130ab237875"


@pytest.fixture
def visualai_embeddings_url(model_url):
    return urljoin(model_url, "imageEmbeddings/")


@pytest.fixture
def visualai_activationmaps_url(model_url):
    return urljoin(model_url, "imageActivationMaps/")


@pytest.fixture(
    params=itertools.product(
        ["aim", "eda", "eda2", "empty", "modeling"],
        [
            TARGET_TYPE.BINARY,
            TARGET_TYPE.MULTICLASS,
            TARGET_TYPE.MULTILABEL,
            TARGET_TYPE.REGRESSION,
        ],
    )
)
def visualai_project(request, project_with_target_json):
    ret = json.loads(project_with_target_json)
    stage, target_type = request.param
    ret["stage"] = stage
    ret["targetType"] = target_type
    return ret


@pytest.fixture
def visualai_image():
    return {"width": 256, "height": 256, "imageId": "5e7e562528513130ab237875"}


@pytest.fixture
def visualai_image_file():
    return bytes((random.randint(0, 255) for i in range(32)))


@pytest.fixture
def visualai_sample(visualai_project):
    target_type_to_eda2_sample_mapping = {
        TARGET_TYPE.BINARY: [
            {
                "imageId": "5e7e562528513130ab237875",
                "height": 256,
                "width": 256,
                "targetValue": "fake",
            },
            {
                "imageId": "5e7e562528513130ab237874",
                "height": 256,
                "width": 256,
                "targetValue": "real",
            },
        ],
        TARGET_TYPE.MULTICLASS: [
            {
                "imageId": "5e7e562528513130ab237875",
                "height": 256,
                "width": 256,
                "targetValue": "fake",
            },
            {
                "imageId": "5e7e562528513130ab237874",
                "height": 256,
                "width": 256,
                "targetValue": "real",
            },
        ],
        TARGET_TYPE.MULTILABEL: [
            {
                "imageId": "5e7e562528513130ab237875",
                "height": 256,
                "width": 256,
                "targetValue": ["foo", "bar"],
            },
            {
                "imageId": "5e7e562528513130ab237874",
                "height": 256,
                "width": 256,
                "targetValue": [],
            },
        ],
        TARGET_TYPE.REGRESSION: [
            {
                "imageId": "5e7e562528513130ab237875",
                "height": 256,
                "width": 256,
                "targetValue": 2.0,
            },
            {"imageId": "5e7e562528513130ab237874", "height": 256, "width": 256, "targetValue": 1},
        ],
    }

    if visualai_project["stage"] in ["eda2", "modeling"]:
        return {
            "next": None,
            "data": target_type_to_eda2_sample_mapping[visualai_project["targetType"]],
            "previous": None,
        }

    else:
        return {
            "next": None,
            "data": [
                {"imageId": "5e7e562528513130ab237875", "height": 256, "width": 256},
                {"imageId": "5e7e562528513130ab237874", "height": 256, "width": 256},
            ],
            "previous": None,
        }


@pytest.fixture
def visualai_duplicate():
    return {
        "count": 0,
        "next": None,
        "data": [
            {"imageId": "5e7e562528513130ab237875", "rowCount": 3},
            {"imageId": "5e7e562528513130ab237874", "rowCount": 7},
        ],
        "previous": None,
    }


@pytest.fixture
def visualai_embeddings():
    return {
        "targetValues": ["fake", "real"],
        "targetBins": None,
        "embeddings": [
            {
                "imageId": "5e7e562b28513130ab23792e",
                "positionX": 0.5935041904449463,
                "positionY": 0.6990952491760254,
                "actualTargetValue": "fake",
            },
            {
                "imageId": "5e7e563528513130ab237a6d",
                "positionX": 0.708981454372406,
                "positionY": 0.6228484511375427,
                "actualTargetValue": "fake",
            },
        ],
    }


@pytest.fixture
def visualai_activationmaps():
    return {
        "activationMapWidth": 56,
        "activationMaps": [
            {
                "imageId": "5eaafb57fcdb565e4d778f29",
                "overlayImageId": "5e7e563528513130ab237a6d",
                "predictedTargetValue": "fake",
                "featureName": "image",
                "actualTargetValue": "fake",
                "imageHeight": 256,
                "imageWidth": 256,
                "activationValues": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            },
            {
                "imageId": "5eaafb58fcdb565e2f7790d0",
                "predictedTargetValue": "real",
                "featureName": "image",
                "actualTargetValue": "real",
                "imageHeight": 256,
                "imageWidth": 256,
                "activationValues": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            },
        ],
        "activationMapHeight": 56,
        "targetBins": None,
        "targetValues": ["fake", "real"],
    }


@pytest.fixture
def visualai_augmentation_list():
    a_param = {"name": "pixels", "currentValue": 10}
    b_param = {"name": "pixels", "currentValue": 12.3}
    trans = {"name": "blur", "params": [a_param]}
    trans2 = {"name": "scale", "params": [b_param]}
    return {
        "id": "5eaafb57fcdb565e4d778f29",
        "name": "my aug list",
        "projectId": "5eaafb57fcdb565e4d778f29",
        "featureName": "image",
        "inUse": False,
        "initialList": True,
        "transformationProbability": 0.5,
        "numberOfNewImages": 1,
        "transformations": [trans, trans2],
    }


@pytest.fixture
def visualai_aug_options():
    a_param = {"name": "pixels", "currentValue": 10}
    b_param = {"name": "pixels", "currentValue": 12.3}
    trans = {"name": "blur", "params": [a_param]}
    trans2 = {"name": "scale", "params": [b_param]}
    return {
        "id": "5eaafb57fcdb565e4d778f29",
        "name": "my aug list",
        "projectId": "5eaafb57fcdb565e4d778f29",
        "minTransformationProbability": 0.1,
        "maxTransformationProbability": 1.0,
        "currentTransformationProbability": 0.5,
        "minNumberOfNewImages": 1,
        "currentNumberOfNewImages": 100,
        "maxNumberOfNewImages": 1796,
        "transformations": [trans, trans2],
    }
